"""
==========================
Data Aggregating Utilities
==========================

This module contains utility functions for aggregating datasets. One primary use 
case for this is combine the results of running sections of the pipeline in an
embarrassingly parallel manner.

Note that it is critical that all data aggregating utility functions are definied
in this module; easylink will not be able to find them otherwise.
"""

import pandas as pd
from loguru import logger


def concatenate_datasets(input_files: list[str], output_filepath: str) -> None:
    """Concatenates multiple datasets into a single one.

    Parameters
    ----------
    input_files
        A list of input file paths to be concatenated.
    output_filepath
        The output filepath.
    """
    logger.info(f"Concatenating {len(input_files)} datasets")
    dfs = [pd.read_parquet(df) for df in input_files]
    df = pd.concat(dfs, ignore_index=True)
    df.to_parquet(output_filepath)



from easylink.pipeline_schema_constants import development, testing

ALLOWED_SCHEMA_PARAMS = {
    "development": development.SCHEMA_PARAMS,
}

TESTING_SCHEMA_PARAMS = {
    "integration": testing.SINGLE_STEP_SCHEMA_PARAMS,
    "combine_bad_topology": testing.BAD_COMBINED_TOPOLOGY_SCHEMA_PARAMS,
    "combine_bad_implementation_names": testing.BAD_COMBINED_TOPOLOGY_SCHEMA_PARAMS,
    "nested_templated_steps": testing.NESTED_TEMPLATED_STEPS_SCHEMA_PARAMS,
    "combine_with_iteration": testing.COMBINE_WITH_ITERATION_SCHEMA_PARAMS,
    "combine_with_iteration_cycle": testing.COMBINE_WITH_ITERATION_SCHEMA_PARAMS,
    "combine_with_extra_node": testing.TRIPLE_STEP_SCHEMA_PARAMS,
}



# mypy: ignore-errors
from layered_config_tree import LayeredConfigTree

from easylink.implementation import Implementation


def test__get_env_vars(mocker):
    mocker.patch(
        "easylink.implementation.load_yaml",
        return_value={
            "test": {"steps": ["this_step"], "env": {"foo": "corge", "spam": "eggs"}}
        },
    )
    implementation = Implementation(
        schema_steps=["this_step"],
        implementation_config=LayeredConfigTree(
            {"name": "test", "configuration": {"foo": "bar", "baz": "qux"}}
        ),
        input_slots=[],
        output_slots=[],
    )
    assert implementation.environment_variables == {
        "foo": "bar",
        "baz": "qux",
        "spam": "eggs",
    }



from pathlib import Path

import pandas as pd
from pyarrow import parquet as pq


def validate_input_file_dummy(filepath: str) -> None:
    extension = Path(filepath).suffix
    if extension == ".parquet":
        output_columns = set(pq.ParquetFile(filepath).schema.names)
    elif extension == ".csv":
        output_columns = set(pd.read_csv(filepath).columns)
    else:
        raise NotImplementedError(
            f"Data file type {extension} is not supported. Convert to Parquet or CSV instead"
        )

    required_columns = {"foo", "bar", "counter"}
    missing_columns = required_columns - output_columns
    if missing_columns:
        raise LookupError(
            f"Data file {filepath} is missing required column(s) {missing_columns}"
        )



"""Unit tests for the various Step classes.

Notes
-----
These unit tests often instantiate a Step object using parameters that would not
actually pass validation in a real-world scenario (i.e. they do not conform to the pipeline
schema). This is intentional because it's easier to flex complexity and edge cases here
rather than try to get full coverage in the e2e tests. It also allows us to test for
future pipeline schema expansion/flexibility in a relative simple manner now.
"""

import re
from typing import Any

import pytest
from layered_config_tree import LayeredConfigTree

from easylink.configuration import Config
from easylink.graph_components import (
    EdgeParams,
    InputSlot,
    InputSlotMapping,
    OutputSlot,
    OutputSlotMapping,
)
from easylink.pipeline_schema_constants.development import NODES
from easylink.step import (
    ChoiceStep,
    EmbarrassinglyParallelStep,
    HierarchicalStep,
    IOStep,
    LoopStep,
    ParallelStep,
    Step,
)
from easylink.utilities.aggregator_utils import concatenate_datasets
from easylink.utilities.splitter_utils import split_data_by_size
from easylink.utilities.validation_utils import validate_input_file_dummy

STEP_KEYS = {step.name: step for step in NODES}


@pytest.fixture
def basic_step_params() -> dict[str, Any]:
    return {
        "step_name": "step_1",
        "input_slots": [
            InputSlot(
                "step_1_main_input",
                "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                validate_input_file_dummy,
            )
        ],
        "output_slots": [OutputSlot("step_1_main_output")],
    }


def test_implementation_node_name(
    basic_step_params: dict[str, Any], default_config: Config
) -> None:
    step = Step(**basic_step_params)
    step.set_configuration_state(default_config["pipeline"]["steps"], {}, {})
    node_name = step.implementation_node_name
    assert node_name == "step_1_python_pandas"

    step.set_parent_step(Step(step_name="foo", name="bar"))
    node_name = step.implementation_node_name
    assert node_name == "bar_step_1_step_1_python_pandas"


@pytest.fixture
def io_step_params() -> dict[str, Any]:
    return {
        "step_name": "io",
        "input_slots": [InputSlot("result", None, validate_input_file_dummy)],
        "output_slots": [OutputSlot("file1")],
    }


def test_io_step_slots(io_step_params: dict[str, Any]) -> None:
    step = IOStep(**io_step_params)
    assert step.name == step.step_name == "io"
    assert step.input_slots == {
        "result": InputSlot("result", None, validate_input_file_dummy)
    }
    assert step.output_slots == {"file1": OutputSlot("file1")}


def test_io_get_implementation_graph(
    io_step_params: dict[str, Any], default_config: Config
) -> None:
    step = IOStep(**io_step_params)
    step._get_config(default_config["pipeline"]["steps"])
    subgraph = step.get_implementation_graph()
    assert list(subgraph.nodes) == ["io"]
    assert list(subgraph.edges) == []


def test_basic_step_slots(basic_step_params: dict[str, Any]) -> None:
    step = Step(**basic_step_params)
    assert step.name == step.step_name == "step_1"
    assert step.input_slots == {
        "step_1_main_input": InputSlot(
            "step_1_main_input",
            "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
            validate_input_file_dummy,
        )
    }
    assert step.output_slots == {"step_1_main_output": OutputSlot("step_1_main_output")}


def test_basic_step_get_implementation_graph(
    basic_step_params: dict[str, Any], default_config: Config
) -> None:
    step = Step(**basic_step_params)
    step.set_configuration_state(default_config["pipeline"]["steps"], {}, {})
    subgraph = step.get_implementation_graph()
    assert list(subgraph.nodes) == ["step_1_python_pandas"]
    assert list(subgraph.edges) == []


@pytest.fixture
def hierarchical_step_params() -> dict[str, Any]:
    return {
        "step_name": "step_4",
        "input_slots": [
            InputSlot(
                "step_4_main_input",
                "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                validate_input_file_dummy,
            )
        ],
        "output_slots": [OutputSlot("step_4_main_output")],
        "nodes": [
            Step(
                "step_4a",
                input_slots=[
                    InputSlot(
                        "step_4a_main_input",
                        "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                        validate_input_file_dummy,
                    )
                ],
                output_slots=[OutputSlot("step_4a_main_output")],
            ),
            Step(
                "step_4b",
                input_slots=[
                    InputSlot(
                        "step_4b_main_input",
                        "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                        validate_input_file_dummy,
                    )
                ],
                output_slots=[OutputSlot("step_4b_main_output")],
            ),
        ],
        "edges": [
            EdgeParams("step_4a", "step_4b", "step_4a_main_output", "step_4b_main_input")
        ],
        "input_slot_mappings": [
            InputSlotMapping("step_4_main_input", "step_4a", "step_4a_main_input")
        ],
        "output_slot_mappings": [
            OutputSlotMapping("step_4_main_output", "step_4b", "step_4b_main_output")
        ],
    }


def test_hierarchical_step_slots(hierarchical_step_params: dict[str, Any]) -> None:
    step = HierarchicalStep(**hierarchical_step_params)
    assert step.name == "step_4"
    assert step.input_slots == {
        "step_4_main_input": InputSlot(
            "step_4_main_input",
            "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
            validate_input_file_dummy,
        )
    }
    assert step.output_slots == {"step_4_main_output": OutputSlot("step_4_main_output")}


def test_hierarchical_step_get_implementation_graph(
    hierarchical_step_params: dict[str, Any]
) -> None:
    step = HierarchicalStep(**hierarchical_step_params)
    pipeline_params = LayeredConfigTree(
        {"step_4": {"implementation": {"name": "step_4_python_pandas", "configuration": {}}}}
    )
    step.set_configuration_state(pipeline_params, {}, {})
    subgraph = step.get_implementation_graph()
    assert list(subgraph.nodes) == ["step_4_python_pandas"]
    assert list(subgraph.edges) == []

    # Test get_implementation_graph for substeps
    pipeline_params = LayeredConfigTree(
        {
            "step_4": {
                "substeps": {
                    "step_4a": {
                        "implementation": {
                            "name": "step_4a_python_pandas",
                            "configuration": {},
                        }
                    },
                    "step_4b": {
                        "implementation": {
                            "name": "step_4b_python_pandas",
                            "configuration": {},
                        }
                    },
                },
            },
        }
    )
    step.set_configuration_state(pipeline_params, {}, {})
    subgraph = step.get_implementation_graph()
    assert list(subgraph.nodes) == [
        "step_4a_python_pandas",
        "step_4b_python_pandas",
    ]
    expected_edges = [
        (
            "step_4a_python_pandas",
            "step_4b_python_pandas",
            {
                "input_slot": InputSlot(
                    "step_4b_main_input",
                    "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    validate_input_file_dummy,
                ),
                "output_slot": OutputSlot("step_4a_main_output"),
                "filepaths": None,
            },
        ),
    ]
    assert len(subgraph.edges) == len(expected_edges)
    for edge in expected_edges:
        assert edge in subgraph.edges(data=True)


@pytest.fixture
def loop_step_params() -> dict[str, Any]:
    return {
        "template_step": HierarchicalStep(
            "step_3",
            input_slots=[
                InputSlot(
                    name="step_3_main_input",
                    env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    validator=validate_input_file_dummy,
                ),
                InputSlot(
                    name="step_3_secondary_input",
                    env_var="DUMMY_CONTAINER_SECONDARY_INPUT_FILE_PATHS",
                    validator=validate_input_file_dummy,
                ),
            ],
            output_slots=[OutputSlot("step_3_main_output")],
            nodes=[
                Step(
                    step_name="step_3a",
                    input_slots=[
                        InputSlot(
                            name="step_3a_main_input",
                            env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                            validator=validate_input_file_dummy,
                        ),
                        InputSlot(
                            name="step_3a_secondary_input",
                            env_var="DUMMY_CONTAINER_SECONDARY_INPUT_FILE_PATHS",
                            validator=validate_input_file_dummy,
                        ),
                    ],
                    output_slots=[OutputSlot("step_3a_main_output")],
                ),
                Step(
                    step_name="step_3b",
                    input_slots=[
                        InputSlot(
                            name="step_3b_main_input",
                            env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                            validator=validate_input_file_dummy,
                        ),
                        InputSlot(
                            name="step_3b_secondary_input",
                            env_var="DUMMY_CONTAINER_SECONDARY_INPUT_FILE_PATHS",
                            validator=validate_input_file_dummy,
                        ),
                    ],
                    output_slots=[OutputSlot("step_3b_main_output")],
                ),
            ],
            edges=[
                EdgeParams(
                    source_node="step_3a",
                    target_node="step_3b",
                    output_slot="step_3a_main_output",
                    input_slot="step_3b_main_input",
                ),
            ],
            input_slot_mappings=[
                InputSlotMapping(
                    parent_slot="step_3_main_input",
                    child_node="step_3a",
                    child_slot="step_3a_main_input",
                ),
                InputSlotMapping(
                    parent_slot="step_3_secondary_input",
                    child_node="step_3a",
                    child_slot="step_3a_secondary_input",
                ),
                InputSlotMapping(
                    parent_slot="step_3_secondary_input",
                    child_node="step_3b",
                    child_slot="step_3b_secondary_input",
                ),
            ],
            output_slot_mappings=[
                OutputSlotMapping(
                    parent_slot="step_3_main_output",
                    child_node="step_3b",
                    child_slot="step_3b_main_output",
                )
            ],
        ),
        "self_edges": [
            EdgeParams(
                source_node="step_3",
                target_node="step_3",
                output_slot="step_3_main_output",
                input_slot="step_3_main_input",
            )
        ],
    }


def test_loop_step_slots(loop_step_params: dict[str, Any]) -> None:
    step = LoopStep(**loop_step_params)
    assert step.name == step.step_name == "step_3"
    assert isinstance(step, LoopStep)
    assert step.input_slots == {
        "step_3_main_input": InputSlot(
            "step_3_main_input",
            "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
            validate_input_file_dummy,
        ),
        "step_3_secondary_input": InputSlot(
            "step_3_secondary_input",
            "DUMMY_CONTAINER_SECONDARY_INPUT_FILE_PATHS",
            validate_input_file_dummy,
        ),
    }
    assert step.output_slots == {"step_3_main_output": OutputSlot("step_3_main_output")}
    assert isinstance(step.template_step, HierarchicalStep)
    assert step.template_step.name == step.name
    assert step.template_step.input_slots == step.input_slots
    assert step.template_step.output_slots == step.output_slots
    assert step.self_edges == [
        EdgeParams(step.name, step.name, "step_3_main_output", "step_3_main_input")
    ]


def test_loop_get_implementation_graph(
    mocker, loop_step_params: dict[str, Any], default_config: Config
) -> None:
    mocker.patch("easylink.implementation.Implementation._load_metadata")
    mocker.patch("easylink.implementation.Implementation.validate", return_value=[])
    step = LoopStep(**loop_step_params)
    step.set_configuration_state(default_config["pipeline"]["steps"], {}, {})
    subgraph = step.get_implementation_graph()
    assert list(subgraph.nodes) == ["step_3_python_pandas"]
    assert list(subgraph.edges) == []

    pipeline_params = LayeredConfigTree(
        {
            "step_3": {
                "iterate": [
                    {
                        "implementation": {
                            "name": "step_3_python_pandas",
                            "configuration": {},
                        },
                    },
                    {
                        "substeps": {
                            "step_3a": {
                                "implementation": {
                                    "name": "step_3a_python_pandas",
                                    "configuration": {},
                                },
                            },
                            "step_3b": {
                                "implementation": {
                                    "name": "step_3b_python_pandas",
                                    "configuration": {},
                                },
                            },
                        },
                    },
                ],
            },
        },
    )
    step.set_configuration_state(pipeline_params, {}, {})
    subgraph = step.get_implementation_graph()
    assert list(subgraph.nodes) == [
        "step_3_loop_1_step_3_python_pandas",
        "step_3_loop_2_step_3a_step_3a_python_pandas",
        "step_3_loop_2_step_3b_step_3b_python_pandas",
    ]
    expected_edges = [
        (
            "step_3_loop_1_step_3_python_pandas",
            "step_3_loop_2_step_3a_step_3a_python_pandas",
            {
                "input_slot": InputSlot(
                    "step_3a_main_input",
                    "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    validate_input_file_dummy,
                ),
                "output_slot": OutputSlot("step_3_main_output"),
                "filepaths": None,
            },
        ),
        (
            "step_3_loop_2_step_3a_step_3a_python_pandas",
            "step_3_loop_2_step_3b_step_3b_python_pandas",
            {
                "input_slot": InputSlot(
                    "step_3b_main_input",
                    "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    validate_input_file_dummy,
                ),
                "output_slot": OutputSlot("step_3a_main_output"),
                "filepaths": None,
            },
        ),
    ]
    assert len(subgraph.edges) == len(expected_edges)
    for edge in expected_edges:
        assert edge in subgraph.edges(data=True)


@pytest.fixture
def parallel_step_params() -> dict[str, Any]:
    return {
        "template_step": HierarchicalStep(
            "step_1",
            input_slots=[
                InputSlot(
                    name="step_1_main_input",
                    env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    validator=validate_input_file_dummy,
                ),
            ],
            output_slots=[OutputSlot("step_1_main_output")],
            nodes=[
                Step(
                    step_name="step_1a",
                    input_slots=[
                        InputSlot(
                            name="step_1a_main_input",
                            env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                            validator=validate_input_file_dummy,
                        ),
                    ],
                    output_slots=[OutputSlot("step_1a_main_output")],
                ),
                Step(
                    step_name="step_1b",
                    input_slots=[
                        InputSlot(
                            name="step_1b_main_input",
                            env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                            validator=validate_input_file_dummy,
                        ),
                    ],
                    output_slots=[OutputSlot("step_1b_main_output")],
                ),
            ],
            edges=[
                EdgeParams(
                    source_node="step_1a",
                    target_node="step_1b",
                    output_slot="step_1a_main_output",
                    input_slot="step_1b_main_input",
                ),
            ],
            input_slot_mappings=[
                InputSlotMapping(
                    parent_slot="step_1_main_input",
                    child_node="step_1a",
                    child_slot="step_1a_main_input",
                )
            ],
            output_slot_mappings=[
                OutputSlotMapping(
                    parent_slot="step_1_main_output",
                    child_node="step_1b",
                    child_slot="step_1b_main_output",
                )
            ],
        ),
    }


def test_parallel_step_slots(parallel_step_params: dict[str, Any]) -> None:
    step = ParallelStep(**parallel_step_params)
    assert step.name == "step_1"
    assert step.input_slots == {
        "step_1_main_input": InputSlot(
            "step_1_main_input",
            "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
            validate_input_file_dummy,
        )
    }
    assert step.output_slots == {"step_1_main_output": OutputSlot("step_1_main_output")}


def test_parallel_step_get_implementation_graph(
    mocker, parallel_step_params: dict[str, Any]
) -> None:
    mocker.patch("easylink.implementation.Implementation._load_metadata")
    mocker.patch("easylink.implementation.Implementation.validate", return_value=[])
    step = ParallelStep(**parallel_step_params)
    pipeline_params = LayeredConfigTree(
        {
            "step_1": {
                "parallel": [
                    {
                        "substeps": {
                            "step_1a": {
                                "implementation": {
                                    "name": "step_1a_python_pandas",
                                },
                            },
                            "step_1b": {
                                "implementation": {
                                    "name": "step_1b_python_pandas",
                                },
                            },
                        },
                        "input_data_file": "input_file_1",
                    },
                    {
                        "substeps": {
                            "step_1a": {
                                "implementation": {
                                    "name": "step_1a_python_pandas",
                                },
                            },
                            "step_1b": {
                                "implementation": {
                                    "name": "step_1b_python_pandas",
                                },
                            },
                        },
                        "input_data_file": "input_file_2",
                    },
                    {
                        "substeps": {
                            "step_1a": {
                                "implementation": {
                                    "name": "step_1a_python_pandas",
                                },
                            },
                            "step_1b": {
                                "implementation": {
                                    "name": "step_1b_python_pandas",
                                },
                            },
                        },
                        "input_data_file": "input_file_3",
                    },
                ],
            },
        },
    )
    step.set_configuration_state(pipeline_params, {}, {})
    subgraph = step.get_implementation_graph()
    assert set(subgraph.nodes) == {
        "step_1_parallel_split_1_step_1a_step_1a_python_pandas",
        "step_1_parallel_split_1_step_1b_step_1b_python_pandas",
        "step_1_parallel_split_2_step_1a_step_1a_python_pandas",
        "step_1_parallel_split_2_step_1b_step_1b_python_pandas",
        "step_1_parallel_split_3_step_1a_step_1a_python_pandas",
        "step_1_parallel_split_3_step_1b_step_1b_python_pandas",
    }
    expected_edges = [
        (
            "step_1_parallel_split_1_step_1a_step_1a_python_pandas",
            "step_1_parallel_split_1_step_1b_step_1b_python_pandas",
            {
                "input_slot": InputSlot(
                    "step_1b_main_input",
                    "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    validate_input_file_dummy,
                ),
                "output_slot": OutputSlot("step_1a_main_output"),
                "filepaths": None,
            },
        ),
        (
            "step_1_parallel_split_2_step_1a_step_1a_python_pandas",
            "step_1_parallel_split_2_step_1b_step_1b_python_pandas",
            {
                "input_slot": InputSlot(
                    "step_1b_main_input",
                    "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    validate_input_file_dummy,
                ),
                "output_slot": OutputSlot("step_1a_main_output"),
                "filepaths": None,
            },
        ),
        (
            "step_1_parallel_split_3_step_1a_step_1a_python_pandas",
            "step_1_parallel_split_3_step_1b_step_1b_python_pandas",
            {
                "input_slot": InputSlot(
                    "step_1b_main_input",
                    "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    validate_input_file_dummy,
                ),
                "output_slot": OutputSlot("step_1a_main_output"),
                "filepaths": None,
            },
        ),
    ]
    assert len(subgraph.edges) == len(expected_edges)
    for edge in expected_edges:
        assert edge in subgraph.edges(data=True)


@pytest.mark.parametrize("step_type", ["parallel", "loop"])
def test_templated_get_implementation_graph_no_multiplicity(
    step_type, parallel_step_params: dict[str, Any], loop_step_params: dict[str, Any], mocker
) -> None:
    """Tests that we can handle TemplatedStep but with no multiplicity."""
    mocker.patch("easylink.implementation.Implementation._load_metadata")
    mocker.patch("easylink.implementation.Implementation.validate", return_value=[])
    if step_type == "parallel":
        step = ParallelStep(**parallel_step_params)
        step_name = "step_1"
    else:  # loop
        step = LoopStep(**loop_step_params)
        step_name = "step_3"
    pipeline_params = LayeredConfigTree(
        {
            step_name: {
                # No multiplicity, i.e. no "parallel" or "loop" key!
                "substeps": {
                    f"{step_name}a": {
                        "implementation": {
                            "name": f"{step_name}a_python_pandas",
                        },
                    },
                    f"{step_name}b": {
                        "implementation": {
                            "name": f"{step_name}b_python_pandas",
                        },
                    },
                },
                "input_data_file": "input_file_1",
            },
        },
    )
    step.set_configuration_state(pipeline_params, {}, {})
    subgraph = step.get_implementation_graph()
    assert set(subgraph.nodes) == {
        f"{step_name}a_python_pandas",
        f"{step_name}b_python_pandas",
    }
    expected_edges = [
        (
            f"{step_name}a_python_pandas",
            f"{step_name}b_python_pandas",
            {
                "input_slot": InputSlot(
                    f"{step_name}b_main_input",
                    "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    validate_input_file_dummy,
                ),
                "output_slot": OutputSlot(f"{step_name}a_main_output"),
                "filepaths": None,
            },
        ),
    ]
    assert len(subgraph.edges) == len(expected_edges)
    for edge in expected_edges:
        assert edge in subgraph.edges(data=True)


@pytest.mark.parametrize("step_type", ["loop", "parallel"])
@pytest.mark.parametrize("single_repeat", [True, False])
def test__duplicate_template_step(
    step_type, single_repeat, loop_step_params, parallel_step_params, mocker
):
    """Test against _duplicate_template_step.

    This is not an exhaustive test due to the complicated nature of testing
    equality of different attributes between two deep copies. For example,
    SomeClass.foo == SomeOtherClass.foo if foo is the same string even though
    they are different objects in memory. This is not the case for methods as
    well as other attribute types.
    """
    mocker.patch("easylink.implementation.Implementation._load_metadata")
    mocker.patch("easylink.implementation.Implementation.validate", return_value=[])
    if step_type == "loop":
        step = LoopStep(**loop_step_params)
        config_key = "iterate"
        step_name = loop_step_params["template_step"].name
    else:  # parallel
        step = ParallelStep(**parallel_step_params)
        config_key = "parallel"
        step_name = parallel_step_params["template_step"].name
    implementation_config = [
        {
            "implementation": {
                "name": f"{step_name}_python_pandas",
            },
        }
    ] * (1 if single_repeat else 2)
    pipeline_params = LayeredConfigTree(
        {
            step_name: {
                config_key: implementation_config,
            },
        },
    )
    step.set_configuration_state(pipeline_params, {}, {})
    step.template_step.set_configuration_state(
        LayeredConfigTree(
            {
                step_name: {
                    "implementation": {
                        "name": f"{step_name}_python_pandas",
                    },
                },
            },
        ),
        {},
        {},
    )

    duplicate_template_step = step._duplicate_template_step()

    special_handle_attrs = [
        "_configuration_state",
        "configuration_state",
        "step_graph",
        "nodes",
    ]
    attrs = [
        attr
        for attr in dir(step.template_step)
        if attr not in special_handle_attrs
        # dunders are too complicated to check
        and not attr.startswith("__")
        # methods are bound to each instance
        and not callable(getattr(step.template_step, attr))
    ]
    for attr in attrs:
        assert getattr(step.template_step, attr) == getattr(duplicate_template_step, attr)

    # Handle the special cases. Just check that they are of the same type and yet
    # not equal (implying that they are bound to different instances)
    for attr in special_handle_attrs:
        assert isinstance(
            getattr(step.template_step, attr),
            type(getattr(duplicate_template_step, attr)),
        )
        assert getattr(step.template_step, attr) != getattr(duplicate_template_step, attr)
        if attr == "nodes":
            assert [node.name for node in step.template_step.nodes] == [
                node.name for node in duplicate_template_step.nodes
            ]


@pytest.fixture
def choice_step_params() -> dict[str, Any]:
    return {
        "step_name": "choice_section",
        "input_slots": [
            InputSlot(
                name="choice_section_main_input",
                env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                validator=validate_input_file_dummy,
            ),
            InputSlot(
                name="choice_section_secondary_input",
                env_var="DUMMY_CONTAINER_SECONDARY_INPUT_FILE_PATHS",
                validator=validate_input_file_dummy,
            ),
        ],
        "output_slots": [OutputSlot("choice_section_main_output")],
        "choices": {
            "simple": {
                "nodes": [
                    HierarchicalStep(
                        step_name="step_4",
                        input_slots=[
                            InputSlot(
                                name="step_4_main_input",
                                env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                                validator=validate_input_file_dummy,
                            ),
                            InputSlot(
                                name="step_4_secondary_input",
                                env_var="DUMMY_CONTAINER_SECONDARY_INPUT_FILE_PATHS",
                                validator=validate_input_file_dummy,
                            ),
                        ],
                        output_slots=[OutputSlot("step_4_main_output")],
                        nodes=[
                            Step(
                                step_name="step_4a",
                                input_slots=[
                                    InputSlot(
                                        name="step_4a_main_input",
                                        env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                                        validator=validate_input_file_dummy,
                                    ),
                                    InputSlot(
                                        name="step_4a_secondary_input",
                                        env_var="DUMMY_CONTAINER_SECONDARY_INPUT_FILE_PATHS",
                                        validator=validate_input_file_dummy,
                                    ),
                                ],
                                output_slots=[OutputSlot("step_4a_main_output")],
                            ),
                            Step(
                                step_name="step_4b",
                                input_slots=[
                                    InputSlot(
                                        name="step_4b_main_input",
                                        env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                                        validator=validate_input_file_dummy,
                                    ),
                                    InputSlot(
                                        name="step_4b_secondary_input",
                                        env_var="DUMMY_CONTAINER_SECONDARY_INPUT_FILE_PATHS",
                                        validator=validate_input_file_dummy,
                                    ),
                                ],
                                output_slots=[OutputSlot("step_4b_main_output")],
                            ),
                        ],
                        edges=[
                            EdgeParams(
                                source_node="step_4a",
                                target_node="step_4b",
                                output_slot="step_4a_main_output",
                                input_slot="step_4b_main_input",
                            ),
                        ],
                        input_slot_mappings=[
                            InputSlotMapping(
                                parent_slot="step_4_main_input",
                                child_node="step_4a",
                                child_slot="step_4a_main_input",
                            ),
                            InputSlotMapping(
                                parent_slot="step_4_secondary_input",
                                child_node="step_4a",
                                child_slot="step_4a_secondary_input",
                            ),
                            InputSlotMapping(
                                parent_slot="step_4_secondary_input",
                                child_node="step_4b",
                                child_slot="step_4b_secondary_input",
                            ),
                        ],
                        output_slot_mappings=[
                            OutputSlotMapping(
                                parent_slot="step_4_main_output",
                                child_node="step_4b",
                                child_slot="step_4b_main_output",
                            ),
                        ],
                    ),
                ],
                "edges": [],
                "input_slot_mappings": [
                    InputSlotMapping(
                        parent_slot="choice_section_main_input",
                        child_node="step_4",
                        child_slot="step_4_main_input",
                    ),
                    InputSlotMapping(
                        parent_slot="choice_section_secondary_input",
                        child_node="step_4",
                        child_slot="step_4_secondary_input",
                    ),
                ],
                "output_slot_mappings": [
                    OutputSlotMapping(
                        parent_slot="choice_section_main_output",
                        child_node="step_4",
                        child_slot="step_4_main_output",
                    ),
                ],
            },
            "complex": {
                "nodes": [
                    Step(
                        step_name="step_5",
                        input_slots=[
                            InputSlot(
                                name="step_5_main_input",
                                env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                                validator=validate_input_file_dummy,
                            ),
                        ],
                        output_slots=[OutputSlot("step_5_main_output")],
                    ),
                    # Add a more complicated (unsupported) loop step to ensure flexibility
                    LoopStep(
                        template_step=Step(
                            step_name="step_6",
                            input_slots=[
                                InputSlot(
                                    name="step_6_main_input",
                                    env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                                    validator=validate_input_file_dummy,
                                ),
                            ],
                            output_slots=[OutputSlot("step_6_main_output")],
                        ),
                        self_edges=[
                            EdgeParams(
                                source_node="step_6",
                                target_node="step_6",
                                output_slot="step_6_main_output",
                                input_slot="step_6_main_input",
                            )
                        ],
                    ),
                ],
                "edges": [
                    EdgeParams(
                        source_node="step_5",
                        target_node="step_6",
                        output_slot="step_5_main_output",
                        input_slot="step_6_main_input",
                    )
                ],
                "input_slot_mappings": [
                    InputSlotMapping(
                        parent_slot="choice_section_main_input",
                        child_node="step_5",
                        child_slot="step_5_main_input",
                    ),
                ],
                "output_slot_mappings": [
                    OutputSlotMapping(
                        parent_slot="choice_section_main_output",
                        child_node="step_6",
                        child_slot="step_6_main_output",
                    ),
                ],
            },
        },
    }


def test_choice_step_slots(choice_step_params: dict[str, Any]) -> None:
    step = ChoiceStep(**choice_step_params)
    assert step.name == "choice_section"
    assert step.input_slots == {
        "choice_section_main_input": InputSlot(
            "choice_section_main_input",
            "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
            validate_input_file_dummy,
        ),
        "choice_section_secondary_input": InputSlot(
            "choice_section_secondary_input",
            "DUMMY_CONTAINER_SECONDARY_INPUT_FILE_PATHS",
            validate_input_file_dummy,
        ),
    }
    assert step.output_slots == {
        "choice_section_main_output": OutputSlot("choice_section_main_output")
    }


def test_simple_choice_step_get_implementation_graph(
    choice_step_params: dict[str, Any]
) -> None:
    step = ChoiceStep(**choice_step_params)

    # Test get_implementation_graph for single step (no substeps)
    pipeline_dict = {
        "choice_section": {
            "type": "simple",
            "step_4": {
                "implementation": {
                    "name": "step_4_python_pandas",
                },
            },
        },
    }
    pipeline_params = LayeredConfigTree(pipeline_dict)
    # Need to validate in order to set the step graph an mappings prior to calling `set_configuration_state`
    step.validate_step(pipeline_dict["choice_section"], {}, {})
    step.set_configuration_state(pipeline_params, {}, {})
    subgraph = step.get_implementation_graph()
    assert list(subgraph.nodes) == ["step_4_python_pandas"]
    assert list(subgraph.edges) == []

    # Test get_implementation_graph for a step with substeps
    pipeline_dict = {
        "choice_section": {
            "type": "simple",
            "step_4": {
                "substeps": {
                    "step_4a": {
                        "implementation": {
                            "name": "step_4a_python_pandas",
                        },
                    },
                    "step_4b": {
                        "implementation": {
                            "name": "step_4b_r",
                        },
                    },
                },
            },
        },
    }
    pipeline_params = LayeredConfigTree(pipeline_dict)
    step.set_configuration_state(pipeline_params, {}, {})
    subgraph = step.get_implementation_graph()
    assert list(subgraph.nodes) == [
        "step_4a_python_pandas",
        "step_4b_r",
    ]
    expected_edges = [
        (
            "step_4a_python_pandas",
            "step_4b_r",
            {
                "input_slot": InputSlot(
                    "step_4b_main_input",
                    "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    validate_input_file_dummy,
                ),
                "output_slot": OutputSlot("step_4a_main_output"),
                "filepaths": None,
            },
        ),
    ]
    assert len(subgraph.edges) == len(expected_edges)
    for edge in expected_edges:
        assert edge in subgraph.edges(data=True)


def test_complex_choice_step_get_implementation_graph(
    choice_step_params: dict[str, Any]
) -> None:
    step = ChoiceStep(**choice_step_params)

    pipeline_dict = {
        "choice_section": {
            "type": "complex",
            "step_5": {
                "implementation": {
                    "name": "step_5_python_pandas",
                },
            },
            "step_6": {
                "iterate": [
                    {
                        "implementation": {
                            "name": "step_6_python_pandas",
                        },
                    },
                    {
                        "implementation": {
                            "name": "step_6_python_pandas",
                        },
                    },
                ],
            },
        },
    }
    pipeline_params = LayeredConfigTree(pipeline_dict)
    # Need to validate in order to set the step graph an mappings prior to calling `set_configuration_state`
    step.validate_step(pipeline_dict["choice_section"], {}, {})
    step.set_configuration_state(pipeline_params, {}, {})
    subgraph = step.get_implementation_graph()
    assert list(subgraph.nodes) == [
        "step_5_python_pandas",
        "step_6_loop_1_step_6_python_pandas",
        "step_6_loop_2_step_6_python_pandas",
    ]
    expected_edges = [
        (
            "step_5_python_pandas",
            "step_6_loop_1_step_6_python_pandas",
            {
                "input_slot": InputSlot(
                    "step_6_main_input",
                    "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    validate_input_file_dummy,
                ),
                "output_slot": OutputSlot("step_5_main_output"),
                "filepaths": None,
            },
        ),
        (
            "step_6_loop_1_step_6_python_pandas",
            "step_6_loop_2_step_6_python_pandas",
            {
                "input_slot": InputSlot(
                    "step_6_main_input",
                    "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    validate_input_file_dummy,
                ),
                "output_slot": OutputSlot("step_6_main_output"),
                "filepaths": None,
            },
        ),
    ]
    assert len(subgraph.edges) == len(expected_edges)
    for edge in expected_edges:
        assert edge in subgraph.edges(data=True)


@pytest.fixture
def embarrassingly_parallel_step_params() -> dict[str, Any]:
    return {
        "step_name": "step_3",
        "input_slots": [
            InputSlot(
                "step_3_main_input",
                "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                validate_input_file_dummy,
                split_data_by_size,
            )
        ],
        "output_slots": [OutputSlot("step_3_main_output", concatenate_datasets)],
    }


def test_embarrassingly_parallel_step_slots(
    embarrassingly_parallel_step_params: dict[str, Any]
) -> None:
    step = EmbarrassinglyParallelStep(**embarrassingly_parallel_step_params)
    assert step.name == "step_3"
    assert step.input_slots == {
        "step_3_main_input": InputSlot(
            "step_3_main_input",
            "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
            validate_input_file_dummy,
            split_data_by_size,
        ),
    }
    assert step.output_slots == {
        "step_3_main_output": OutputSlot("step_3_main_output", concatenate_datasets)
    }


def test_embarrassingly_parallel_step_get_implementation_graph(
    embarrassingly_parallel_step_params: dict[str, Any]
) -> None:
    step = EmbarrassinglyParallelStep(**embarrassingly_parallel_step_params)
    pipeline_params = LayeredConfigTree(
        {"step_3": {"implementation": {"name": "step_3_python_pandas", "configuration": {}}}}
    )
    step.set_configuration_state(pipeline_params, {}, {})
    subgraph = step.get_implementation_graph()
    assert list(subgraph.nodes) == ["step_3_python_pandas"]
    assert list(subgraph.edges) == []


@pytest.mark.parametrize(
    "input_slots, output_slots, expected_error_msg",
    [
        (
            [
                InputSlot(
                    "main_input",
                    "FOO",
                    validate_input_file_dummy,
                ),
            ],
            [OutputSlot("main_output", concatenate_datasets)],
            "does not have any input slots with a splitter method assigned",
        ),
        (
            [
                InputSlot(
                    "main_input",
                    "FOO",
                    validate_input_file_dummy,
                    split_data_by_size,
                ),
                InputSlot(
                    "secondary_input",
                    "BAR",
                    validate_input_file_dummy,
                    split_data_by_size,
                ),
            ],
            [OutputSlot("main_output", concatenate_datasets)],
            "has multiple input slots with splitter methods assigned",
        ),
        (
            [
                InputSlot(
                    "main_input",
                    "FOO",
                    validate_input_file_dummy,
                    split_data_by_size,
                ),
            ],
            [
                OutputSlot("main_output", concatenate_datasets),
                OutputSlot("secondary_output"),
            ],
            "has output slots without aggregator methods",
        ),
        (
            [
                InputSlot(
                    "main_input",
                    "FOO",
                    validate_input_file_dummy,
                ),
            ],
            [
                OutputSlot("main_output", concatenate_datasets),
                OutputSlot("secondary_output"),
            ],
            [
                "does not have any input slots with a splitter method assigned",
                "has output slots without aggregator methods",
            ],
        ),
        (
            [
                InputSlot(
                    "main_input",
                    "FOO",
                    validate_input_file_dummy,
                    split_data_by_size,
                ),
                InputSlot(
                    "secondary_input",
                    "BAR",
                    validate_input_file_dummy,
                    split_data_by_size,
                ),
            ],
            [
                OutputSlot("main_output", concatenate_datasets),
                OutputSlot("secondary_output"),
            ],
            [
                "has multiple input slots with splitter methods assigned",
                "has output slots without aggregator methods",
            ],
        ),
    ],
    ids=[
        "no_splitter",
        "multiple_splitters",
        "missing_aggregators",
        "no_splitter_missing_aggregators",
        "multiple_splitters_missing_aggregators",
    ],
)
def test_embarrassingly_parallel_step__validation(
    input_slots: list[InputSlot],
    output_slots: list[OutputSlot],
    expected_error_msg: str | list[str],
):
    step_params = {
        "step_name": "step",
        "input_slots": input_slots,
        "output_slots": output_slots,
    }
    with pytest.raises(ValueError) as error:
        EmbarrassinglyParallelStep(**step_params)

    error_msg = str(error.value)

    expected = (
        [expected_error_msg] if isinstance(expected_error_msg, str) else expected_error_msg
    )
    for msg in expected:
        assert msg in error_msg



"""
===============
Implementations
===============

This module is responsible for defining the abstractions that represent actual
implementations of steps in a pipeline. Typically, these abstractions contain
information about what container to run for a given step and other related details.

"""

from collections.abc import Iterable
from pathlib import Path

from layered_config_tree import LayeredConfigTree

from easylink.graph_components import InputSlot, OutputSlot
from easylink.utilities import paths
from easylink.utilities.data_utils import load_yaml


class Implementation:
    """A representation of an actual container that will be executed for a :class:`~easylink.step.Step`.

    ``Implementations`` exist at a lower level than ``Steps``. This class contains
    information about what container to use, what environment variables to set
    inside the container, and some metadata about the container.

    Parameters
    ----------
    schema_steps
        The user-requested ``Step`` names for which this ``Implementation`` is
        expected to implement.
    implementation_config
        The configuration details required to run the relevant container.
    input_slots
        All required :class:`InputSlots<easylink.graph_components.InputSlot>`.
    output_slots
        All required :class:`OutputSlots<easylink.graph_components.OutputSlot>`.
    """

    def __init__(
        self,
        schema_steps: list[str],
        implementation_config: LayeredConfigTree,
        input_slots: Iterable["InputSlot"] = (),
        output_slots: Iterable["OutputSlot"] = (),
        is_embarrassingly_parallel: bool = False,
    ):
        self.name = implementation_config.name
        """The name of this ``Implementation``."""
        self.input_slots = {slot.name: slot for slot in input_slots}
        """A mapping of ``InputSlot`` names to their instances."""
        self.output_slots = {slot.name: slot for slot in output_slots}
        """A mapping of ``OutputSlot`` names to their instances."""
        self._metadata = self._load_metadata()
        self.environment_variables = self._get_env_vars(implementation_config)
        """A mapping of environment variables to set."""
        self.metadata_steps = self._metadata["steps"]
        """The names of the specific ``Steps`` for which this ``Implementation`` 
        has been designed to implement."""
        self.schema_steps = schema_steps
        """The names of the specific ``Steps`` that the user has requested to be
        implemented by this particular ``Implementation``."""
        self.requires_spark = self._metadata.get("requires_spark", False)
        """Whether this ``Implementation`` requires a Spark environment."""
        self.is_embarrassingly_parallel = is_embarrassingly_parallel

    def __repr__(self) -> str:
        return f"Implementation.{self.name}"

    def validate(self) -> list[str]:
        """Validates individual ``Implementation`` instances.

        Returns
        -------
            A list of logs containing any validation errors. Each item in the list
            is a distinct message about a particular validation error (e.g. if a
            required container does not exist).

        Notes
        -----
        This is intended to be run from :meth:`easylink.pipeline.Pipeline._validate`.
        """
        logs = []
        logs = self._validate_expected_steps(logs)
        logs = self._validate_container_exists(logs)
        return logs

    ##################
    # Helper methods #
    ##################

    def _load_metadata(self) -> dict[str, str]:
        """Loads the relevant implementation metadata."""
        metadata = load_yaml(paths.IMPLEMENTATION_METADATA)
        return metadata[self.name]

    def _validate_expected_steps(self, logs: list[str]) -> list[str]:
        """Validates that the ``Implementation`` is responsible for the correct steps."""
        if not set(self.schema_steps) == set(self.metadata_steps):
            logs.append(
                f"Pipeline configuration nodes {self.schema_steps} do not match "
                f"metadata steps {self.metadata_steps}."
            )
        return logs

    def _validate_container_exists(self, logs: list[str]) -> list[str]:
        """Validates that the container to run exists."""
        err_str = f"Container '{self.singularity_image_path}' does not exist."
        if not Path(self.singularity_image_path).exists():
            logs.append(err_str)
        return logs

    def _get_env_vars(self, implementation_config: LayeredConfigTree) -> dict[str, str]:
        """Gets the relevant environment variables."""
        env_vars = self._metadata.get("env", {})
        env_vars.update(implementation_config.get("configuration", {}))
        return env_vars

    @property
    def singularity_image_path(self) -> str:
        """The path to the required Singularity image."""
        return self._metadata["image_path"]

    @property
    def script_cmd(self) -> str:
        """The command to run inside of the container."""
        return self._metadata["script_cmd"]

    @property
    def outputs(self) -> dict[str, list[str]]:
        """The expected output metadata."""
        return self._metadata["outputs"]


class NullImplementation:
    """A partial :class:`Implementation` interface when no container is needed to run.

    The primary use case for this class is when adding an
    :class:`~easylink.step.IOStep` - which does not have a corresponding
    ``Implementation`` - to an :class:`~easylink.graph_components.ImplementationGraph`
    since adding any new node requires an object with :class:`~easylink.graph_components.InputSlot`
    and :class:`~easylink.graph_components.OutputSlot` names.

    Parameters
    ----------
    name
        The name of this ``NullImplementation``.
    input_slots
        All required ``InputSlots``.
    output_slots
        All required ``OutputSlots``.
    """

    def __init__(
        self,
        name: str,
        input_slots: Iterable["InputSlot"] = (),
        output_slots: Iterable["OutputSlot"] = (),
    ):
        self.name = name
        """The name of this ``NullImplementation``."""
        self.input_slots = {slot.name: slot for slot in input_slots}
        """A mapping of ``InputSlot`` names to their instances."""
        self.output_slots = {slot.name: slot for slot in output_slots}
        """A mapping of ``OutputSlot`` names to their instances."""
        self.schema_steps = [self.name]
        """The requested :class:`~easylink.step.Step` names this ``NullImplementation`` implements."""
        self.combined_name = None
        """The name of the combined implementation of which ``NullImplementation`` 
        is a constituent. This is definitionally None."""


class PartialImplementation:
    """One part of a combined implementation that spans multiple :class:`Steps<easylink.step.Step>`.

    A ``PartialImplementation`` is what is initially added to the
    :class:`~easylink.graph_components.ImplementationGraph` when a so-called
    "combined implementation" is used (i.e. an :class:`Implementation` that spans
    multiple ``Steps``). We initially add a node for *each* ``Step``, which has as
    its ``implementation`` attribute a ``PartialImplementation``. Such a graph is not
    yet fit to run. When we make our second pass through, after the flat (non-hierarchical)
    :class:`~easylink.pipeline_graph.PipelineGraph` has been created, we find the
    set of ``PartialImplementation`` nodes corresponding to each combined implementation
    and replace them with a single node with a true ``Implementation`` representing
    the combined implementation.

    Parameters
    ----------
    combined_name
        The name of the combined implementation of which this ``PartialImplementation``
        is a part.
    schema_step
        The requested ``Step`` name that this ``PartialImplementation`` partially
        implements.
    input_slots
        The :class:`InputSlots<easylink.graph_components.InputSlot>` for this ``PartialImplementation``.
    output_slots
        The :class:`OutputSlots<easylink.graph_components.OutputSlot>` for this ``PartialImplementation``.

    """

    def __init__(
        self,
        combined_name: str,
        schema_step: str,
        input_slots: Iterable["InputSlot"] = (),
        output_slots: Iterable["OutputSlot"] = (),
    ):
        self.combined_name = combined_name
        """The name of the combined implementation of which this ``PartialImplementation``
        is a part."""
        self.schema_step = schema_step
        """The requested ``Step`` name that this ``PartialImplementation`` partially 
        implements."""
        self.input_slots = {slot.name: slot for slot in input_slots}
        """A mapping of ``InputSlot`` names to their instances."""
        self.output_slots = {slot.name: slot for slot in output_slots}
        """A mapping of ``OutputSlot`` names to their instances."""



# mypy: ignore-errors
from pathlib import Path

import pytest
from _pytest.logging import LogCaptureFixture
from loguru import logger

SPECIFICATIONS_DIR = Path("tests/specifications")
RESULTS_DIR = "/mnt/team/simulation_science/priv/engineering/tests/output/"


def pytest_addoption(parser):
    parser.addoption("--runslow", action="store_true", default=False, help="run slow tests")


def pytest_configure(config):
    config.addinivalue_line("markers", "slow: mark test as slow to run")


def pytest_collection_modifyitems(config, items):
    if config.getoption("--runslow"):
        # --runslow given in cli: do not skip slow tests
        return
    skip_slow = pytest.mark.skip(reason="need --runslow option to run")
    for item in items:
        if "slow" in item.keywords:
            item.add_marker(skip_slow)


@pytest.fixture
def caplog(caplog: LogCaptureFixture):
    handler_id = logger.add(
        caplog.handler,
        format="{message}",
        level=0,
        filter=lambda record: record["level"].no >= caplog.handler.level,
        enqueue=False,  # Set to 'True' if your test is spawning child processes.
    )
    yield caplog
    logger.remove(handler_id)



# mypy: ignore-errors
import hashlib
import os
import subprocess
import sys
import tempfile
from pathlib import Path
from pprint import pprint

import pytest

from easylink.utilities.data_utils import load_yaml
from easylink.utilities.general_utils import is_on_slurm
from tests.conftest import RESULTS_DIR, SPECIFICATIONS_DIR

RESULT_CHECKSUM = "d56f3d12c0c9ca89b6ce0f42810bb79cd16772535f73807722e5430819b6bc08"


@pytest.mark.slow
@pytest.mark.skipif(
    not is_on_slurm(),
    reason="Must be on slurm to run this test.",
)
@pytest.mark.parametrize(
    "pipeline_specification, input_data, computing_environment",
    [
        # slurm
        (
            "e2e/pipeline.yaml",
            "common/input_data.yaml",
            "e2e/environment_slurm.yaml",
        ),
        # local
        (
            "e2e/pipeline.yaml",
            "common/input_data.yaml",
            "common/environment_local.yaml",
        ),
    ],
)
def test_easylink_run(pipeline_specification, input_data, computing_environment, capsys):
    """Tests the 'easylink run' command

    Notes
    -----
    We use various print statements in this test because they show up in the
    Jenkins logs.
    """
    # Create a temporary directory to store results. We cannot use pytest's tmp_path fixture
    # because other nodes do not have access to it. Also, do not use a context manager
    # (i.e. tempfile.TemporaryDirectory) because it's too difficult to debug when the test
    # fails b/c the dir gets deleted.
    results_dir = tempfile.mkdtemp(dir=RESULTS_DIR)
    # give the tmpdir the same permissions as the parent directory so that
    # cluster jobs can write to it
    os.chmod(results_dir, os.stat(RESULTS_DIR).st_mode)
    results_dir = Path(results_dir)
    with capsys.disabled():  # disabled so we can monitor job submissions
        print(
            "\n\n*** RUNNING TEST ***\n"
            f"[{pipeline_specification}, {input_data}, {computing_environment}]\n"
        )

        cmd = (
            "easylink run "
            f"-p {SPECIFICATIONS_DIR / pipeline_specification} "
            f"-i {SPECIFICATIONS_DIR / input_data} "
            f"-e {SPECIFICATIONS_DIR / computing_environment} "
            f"-o {str(results_dir)} "
            "--no-timestamp"
        )
        subprocess.run(
            cmd,
            shell=True,
            stdout=sys.stdout,
            stderr=sys.stderr,
            check=True,
        )
        final_output = results_dir / "result.parquet"
        assert final_output.exists()
        # Check that the results file checksum matches the expected value
        with open(final_output, "rb") as f:
            actual_checksum = hashlib.sha256(f.read()).hexdigest()
        assert actual_checksum == RESULT_CHECKSUM

        assert (results_dir / Path(pipeline_specification).name).exists()
        assert (results_dir / Path(input_data).name).exists()
        assert (results_dir / Path(computing_environment).name).exists()

        # Check that implementation configuration worked
        diagnostics_dir = results_dir / "diagnostics"
        assert (
            load_yaml(diagnostics_dir / "step_1_python_pandas" / "diagnostics.yaml")[
                "increment"
            ]
            == 1
        )
        assert (
            load_yaml(diagnostics_dir / "step_2_python_pyspark" / "diagnostics.yaml")[
                "increment"
            ]
            == 100
        )
        assert (
            load_yaml(diagnostics_dir / "step_3_python_pandas" / "diagnostics.yaml")[
                "increment"
            ]
            == 702
        )
        assert (
            load_yaml(diagnostics_dir / "step_4_r" / "diagnostics.yaml")["increment"] == 912
        )
        # If it made it through all this, print some diagnostics and delete the results_dir
        final_diagnostics = load_yaml(
            sorted([d for d in diagnostics_dir.iterdir() if d.is_dir()])[-1]
            / "diagnostics.yaml"
        )
        print("\nFinal diagnostics:\n")
        pprint(final_diagnostics)
        # sleep briefly before removing tmpdir b/c sometimes one lingers
        os.system(f"sleep 1 && rm -rf {results_dir}")
        print(
            "\n\n*** END OF TEST ***\n"
            f"[{pipeline_specification}, {input_data}, {computing_environment}]\n"
        )



import re
from pathlib import Path

import pytest

from easylink.configuration import Config
from easylink.graph_components import InputSlot, OutputSlot
from easylink.pipeline_graph import PipelineGraph
from easylink.pipeline_schema import PipelineSchema
from easylink.pipeline_schema_constants import TESTING_SCHEMA_PARAMS
from easylink.utilities.aggregator_utils import concatenate_datasets
from easylink.utilities.data_utils import load_yaml
from easylink.utilities.splitter_utils import split_data_by_size
from easylink.utilities.validation_utils import validate_input_file_dummy


def test__create_graph(default_config: Config, test_dir: str) -> None:
    pipeline_graph = PipelineGraph(default_config)
    expected_nodes = {
        "input_data",
        "step_1_python_pandas",
        "step_2_python_pandas",
        "step_3_python_pandas",
        "step_4_python_pandas",
        "results",
    }
    expected_edges = {
        ("input_data", "step_1_python_pandas"): {
            "input_slot_name": "step_1_main_input",
            "output_slot_name": "all",
            "validator": validate_input_file_dummy,
            "env_var": "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
            "filepaths": (
                Path(f"{test_dir}/input_data1/file1.csv"),
                Path(f"{test_dir}/input_data2/file2.csv"),
            ),
        },
        ("input_data", "step_4_python_pandas"): {
            "input_slot_name": "step_4_secondary_input",
            "output_slot_name": "all",
            "validator": validate_input_file_dummy,
            "env_var": "DUMMY_CONTAINER_SECONDARY_INPUT_FILE_PATHS",
            "filepaths": (
                Path(f"{test_dir}/input_data1/file1.csv"),
                Path(f"{test_dir}/input_data2/file2.csv"),
            ),
        },
        ("step_1_python_pandas", "step_2_python_pandas"): {
            "input_slot_name": "step_2_main_input",
            "output_slot_name": "step_1_main_output",
            "validator": validate_input_file_dummy,
            "env_var": "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
            "filepaths": (Path("intermediate/step_1_python_pandas/result.parquet"),),
        },
        ("step_2_python_pandas", "step_3_python_pandas"): {
            "input_slot_name": "step_3_main_input",
            "output_slot_name": "step_2_main_output",
            "validator": validate_input_file_dummy,
            "env_var": "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
            "filepaths": (Path("intermediate/step_2_python_pandas/result.parquet"),),
        },
        ("step_3_python_pandas", "step_4_python_pandas"): {
            "input_slot_name": "step_4_main_input",
            "output_slot_name": "step_3_main_output",
            "validator": validate_input_file_dummy,
            "env_var": "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
            "filepaths": (Path("intermediate/step_3_python_pandas/result.parquet"),),
        },
        ("step_4_python_pandas", "results"): {
            "input_slot_name": "result",
            "output_slot_name": "step_4_main_output",
            "validator": validate_input_file_dummy,
            "env_var": None,
            "filepaths": (Path("intermediate/step_4_python_pandas/result.parquet"),),
        },
    }
    check_nodes_and_edges(pipeline_graph, expected_nodes, expected_edges)


def test_implementations(default_config: Config) -> None:
    pipeline_graph = PipelineGraph(default_config)
    implementation_names = [
        implementation.name for implementation in pipeline_graph.implementations
    ]
    expected_names = [
        "step_1_python_pandas",
        "step_2_python_pandas",
        "step_3_python_pandas",
        "step_4_python_pandas",
    ]
    assert implementation_names == expected_names
    assert pipeline_graph.implementation_nodes == expected_names


@pytest.mark.parametrize(
    "slot_type, is_duplicated, slot_tuples",
    [
        (
            "input_slot",
            False,
            {
                ("step_a", InputSlot("foo", env_var="bar", validator=None)),
                ("step_b", InputSlot("baz", env_var="spam", validator=None)),
            },
        ),
        (
            "input_slot",
            True,
            {
                ("step_a", InputSlot("foo", env_var="bar", validator=None)),
                ("step_b", InputSlot("foo", env_var="spam", validator=None)),
            },
        ),
        (
            "input_slot",
            True,
            {
                ("step_a", InputSlot("foo", env_var="bar", validator=None)),
                ("step_b", InputSlot("baz", env_var="bar", validator=None)),
            },
        ),
        (
            "output_slot",
            False,
            {
                ("step_a", OutputSlot("foo")),
                ("step_b", OutputSlot("baz")),
            },
        ),
        (
            "output_slot",
            True,
            {
                ("step_a", OutputSlot("foo")),
                ("step_b", OutputSlot("foo")),
            },
        ),
    ],
)
def test__get_duplicate_slots(slot_tuples, slot_type, is_duplicated) -> None:
    duplicates = PipelineGraph._get_duplicate_slots(slot_tuples, slot_type)
    if is_duplicated:
        assert duplicates == slot_tuples
    else:
        assert not duplicates


def test_update_slot_filepaths(default_config: Config, test_dir: str) -> None:
    pipeline_graph = PipelineGraph(default_config)
    pipeline_graph._update_slot_filepaths(default_config)
    expected_filepaths = {
        ("input_data", "step_1_python_pandas"): (
            str(Path(f"{test_dir}/input_data1/file1.csv")),
            str(Path(f"{test_dir}/input_data2/file2.csv")),
        ),
        ("input_data", "step_4_python_pandas"): (
            str(Path(f"{test_dir}/input_data1/file1.csv")),
            str(Path(f"{test_dir}/input_data2/file2.csv")),
        ),
        ("step_1_python_pandas", "step_2_python_pandas"): (
            str(Path("intermediate/step_1_python_pandas/result.parquet")),
        ),
        ("step_2_python_pandas", "step_3_python_pandas"): (
            str(Path("intermediate/step_2_python_pandas/result.parquet")),
        ),
        ("step_3_python_pandas", "step_4_python_pandas"): (
            str(Path("intermediate/step_3_python_pandas/result.parquet")),
        ),
        ("step_4_python_pandas", "results"): (
            str(Path("intermediate/step_4_python_pandas/result.parquet")),
        ),
    }
    for source, sink, edge_attrs in pipeline_graph.edges(data=True):
        assert edge_attrs["filepaths"] == expected_filepaths[(source, sink)]


def test_get_io_slots(default_config: Config, test_dir: str) -> None:
    expected = {
        "step_1_python_pandas": {
            "input": {
                "step_1_main_input": {
                    "env_var": "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    "validator": validate_input_file_dummy,
                    "files": [
                        Path(f"{test_dir}/input_data1/file1.csv"),
                        Path(f"{test_dir}/input_data2/file2.csv"),
                    ],
                    "splitter": None,
                },
            },
            "output": {
                "step_1_main_output": {
                    "filepaths": [Path("intermediate/step_1_python_pandas/result.parquet")],
                    "aggregator": None,
                },
            },
        },
        "step_2_python_pandas": {
            "input": {
                "step_2_main_input": {
                    "env_var": "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    "validator": validate_input_file_dummy,
                    "files": [Path("intermediate/step_1_python_pandas/result.parquet")],
                    "splitter": None,
                },
            },
            "output": {
                "step_2_main_output": {
                    "filepaths": [Path("intermediate/step_2_python_pandas/result.parquet")],
                    "aggregator": None,
                },
            },
        },
        "step_3_python_pandas": {
            "input": {
                "step_3_main_input": {
                    "env_var": "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    "validator": validate_input_file_dummy,
                    "files": [Path("intermediate/step_2_python_pandas/result.parquet")],
                    "splitter": split_data_by_size,
                },
            },
            "output": {
                "step_3_main_output": {
                    "filepaths": [Path("intermediate/step_3_python_pandas/result.parquet")],
                    "aggregator": concatenate_datasets,
                },
            },
        },
        "step_4_python_pandas": {
            "input": {
                "step_4_main_input": {
                    "env_var": "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    "validator": validate_input_file_dummy,
                    "files": [Path("intermediate/step_3_python_pandas/result.parquet")],
                    "splitter": None,
                },
            },
            "output": {
                "step_4_main_output": {
                    "filepaths": [Path("intermediate/step_4_python_pandas/result.parquet")],
                    "aggregator": None,
                },
            },
        },
    }
    pipeline_graph = PipelineGraph(default_config)
    for node, expected_slots in expected.items():
        input_slots, output_slots = pipeline_graph.get_io_slot_attributes(node)
        for slot_name, expected_slot in expected_slots["input"].items():
            slot = input_slots[slot_name]
            assert len(slot) == 4
            assert slot["env_var"] == expected_slot["env_var"]
            assert slot["validator"] == expected_slot["validator"]
            assert slot["filepaths"] == [str(file) for file in expected_slot["files"]]
            assert slot["splitter"] == expected_slot["splitter"]
        for slot_name, expected_slot in expected_slots["output"].items():
            slot = output_slots[slot_name]
            assert len(slot) == 2
            assert slot["filepaths"] == [str(file) for file in expected_slot["filepaths"]]
            assert slot["aggregator"] == expected_slot["aggregator"]


def test_condense_input_slots() -> None:
    input_slots = [
        InputSlot(
            "step_1_main_input",
            "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
            validate_input_file_dummy,
        ),
        InputSlot(
            "step_1_main_input",
            "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
            validate_input_file_dummy,
        ),
        InputSlot(
            "step_2_main_input",
            "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
            validate_input_file_dummy,
        ),
        InputSlot(
            "step_2_main_input",
            "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
            validate_input_file_dummy,
        ),
    ]
    filepaths_by_slot = [
        ["file1", "file2"],
        ["file3", "file4"],
        ["file5", "file6"],
        ["file7", "file8"],
    ]
    expected = {
        "step_1_main_input": {
            "env_var": "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
            "validator": validate_input_file_dummy,
            "filepaths": ["file1", "file2", "file3", "file4"],
        },
        "step_2_main_input": {
            "env_var": "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
            "validator": validate_input_file_dummy,
            "filepaths": ["file5", "file6", "file7", "file8"],
        },
    }
    condensed_slots = PipelineGraph._deduplicate_input_slots(input_slots, filepaths_by_slot)
    for slot_name, expected_slot in expected.items():
        slot = condensed_slots[slot_name]
        assert slot["env_var"] == expected_slot["env_var"]
        assert slot["validator"] == expected_slot["validator"]
        assert slot["filepaths"] == expected_slot["filepaths"]


def test_condense_input_slots_duplicate_slots_raises() -> None:
    input_slots = [
        InputSlot(
            "step_1_main_input",
            "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
            validate_input_file_dummy,
        ),
        InputSlot(
            "step_1_main_input",
            "DUMMY_CONTAINER_SECONDARY_INPUT_FILE_PATHS",
            validate_input_file_dummy,
        ),
    ]
    filepaths_by_slot = [
        ["file1", "file2"],
        ["file3", "file4"],
    ]
    with pytest.raises(ValueError):
        PipelineGraph._deduplicate_input_slots(input_slots, filepaths_by_slot)


def test_condense_input_slots_duplicate_slots_raises() -> None:
    input_slots = [
        InputSlot(
            "step_1_main_input",
            "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
            validate_input_file_dummy,
        ),
        InputSlot(
            "step_1_main_input", "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS", lambda x: None
        ),
    ]
    filepaths_by_slot = [
        ["file1", "file2"],
        ["file3", "file4"],
    ]
    with pytest.raises(ValueError):
        PipelineGraph._deduplicate_input_slots(input_slots, filepaths_by_slot)


def test_get_input_output_files(default_config: Config, test_dir: str) -> None:
    expected = {
        "step_1_python_pandas": (
            [
                f"{test_dir}/input_data1/file1.csv",
                f"{test_dir}/input_data2/file2.csv",
            ],
            ["intermediate/step_1_python_pandas/result.parquet"],
        ),
        "step_2_python_pandas": (
            ["intermediate/step_1_python_pandas/result.parquet"],
            ["intermediate/step_2_python_pandas/result.parquet"],
        ),
        "step_3_python_pandas": (
            ["intermediate/step_2_python_pandas/result.parquet"],
            ["intermediate/step_3_python_pandas/result.parquet"],
        ),
        "step_4_python_pandas": (
            [
                f"{test_dir}/input_data1/file1.csv",
                f"{test_dir}/input_data2/file2.csv",
                "intermediate/step_3_python_pandas/result.parquet",
            ],
            ["intermediate/step_4_python_pandas/result.parquet"],
        ),
    }
    pipeline_graph = PipelineGraph(default_config)
    for node, (expected_input_files, expected_output_files) in expected.items():
        input_files, output_files = pipeline_graph.get_io_filepaths(node)
        assert input_files == expected_input_files
        assert output_files == expected_output_files


@pytest.mark.parametrize("requires_spark", [True, False])
def test_spark_is_required(default_config_params, requires_spark):
    config_params = default_config_params
    if requires_spark:
        # Change step 1's implementation to python_pyspark
        config_params["pipeline"]["steps"]["step_1"]["implementation"][
            "name"
        ] = "step_1_python_pyspark"
    config = Config(config_params)
    pipeline_graph = PipelineGraph(config)
    assert pipeline_graph.spark_is_required == requires_spark


def test_get_whether_embarrassingly_parallel(default_config_params):
    config_params = default_config_params
    config = Config(config_params)
    pipeline_graph = PipelineGraph(config)
    nodes = [node for node in pipeline_graph.nodes if node not in ["input_data", "results"]]
    for node in nodes:
        if node == "step_3_python_pandas":
            assert pipeline_graph.get_whether_embarrassingly_parallel(node)
        else:
            assert not pipeline_graph.get_whether_embarrassingly_parallel(node)


@pytest.mark.parametrize("any_embarrassingly_parallel", [True, False])
def test_any_embarrassingly_parallel(default_config_params, any_embarrassingly_parallel):
    config = Config(default_config_params)
    pipeline_graph = PipelineGraph(config, freeze=False)
    if not any_embarrassingly_parallel:
        pipeline_graph.remove_node("step_3_python_pandas")
    assert pipeline_graph.any_embarrassingly_parallel == any_embarrassingly_parallel


def test_merge_combined_implementations(
    default_config_params, test_dir, unit_test_specifications_dir
) -> None:
    config_params = default_config_params
    # combine steps 1 and 2
    config_params["pipeline"] = load_yaml(
        f"{unit_test_specifications_dir}/pipeline_combine_two_steps.yaml"
    )
    pipeline_graph = PipelineGraph(Config(config_params))
    expected_nodes = {
        "input_data",
        "step_1_2",
        "step_3_python_pandas",
        "step_4_python_pandas",
        "results",
    }
    expected_edges = {
        ("input_data", "step_1_2"): {
            "input_slot_name": "step_1_main_input",
            "output_slot_name": "all",
            "validator": validate_input_file_dummy,
            "env_var": "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
            "filepaths": (
                Path(f"{test_dir}/input_data1/file1.csv"),
                Path(f"{test_dir}/input_data2/file2.csv"),
            ),
        },
        ("input_data", "step_4_python_pandas"): {
            "input_slot_name": "step_4_secondary_input",
            "output_slot_name": "all",
            "validator": validate_input_file_dummy,
            "env_var": "DUMMY_CONTAINER_SECONDARY_INPUT_FILE_PATHS",
            "filepaths": (
                Path(f"{test_dir}/input_data1/file1.csv"),
                Path(f"{test_dir}/input_data2/file2.csv"),
            ),
        },
        ("step_1_2", "step_3_python_pandas"): {
            "input_slot_name": "step_3_main_input",
            "output_slot_name": "step_2_main_output",
            "validator": validate_input_file_dummy,
            "env_var": "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
            "filepaths": (Path("intermediate/step_1_2/result.parquet"),),
        },
        ("step_3_python_pandas", "step_4_python_pandas"): {
            "input_slot_name": "step_4_main_input",
            "output_slot_name": "step_3_main_output",
            "validator": validate_input_file_dummy,
            "env_var": "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
            "filepaths": (Path("intermediate/step_3_python_pandas/result.parquet"),),
        },
        ("step_4_python_pandas", "results"): {
            "input_slot_name": "result",
            "output_slot_name": "step_4_main_output",
            "validator": validate_input_file_dummy,
            "env_var": None,
            "filepaths": (Path("intermediate/step_4_python_pandas/result.parquet"),),
        },
    }
    check_nodes_and_edges(pipeline_graph, expected_nodes, expected_edges)


def test_merge_combined_implementations_iteration(
    default_config_params, test_dir, unit_test_specifications_dir
) -> None:
    config_params = default_config_params
    config_params["pipeline"] = load_yaml(
        f"{unit_test_specifications_dir}/pipeline_combine_with_iteration.yaml"
    )
    schema = PipelineSchema(
        "combine_with_iteration", *TESTING_SCHEMA_PARAMS["combine_with_iteration"]
    )
    config = Config(config_params, schema)
    pipeline_graph = PipelineGraph(config)

    expected_nodes = {
        "input_data",
        "step_1_loop_1_step_1_python_pandas",
        "step_1_2",
        "results",
    }
    expected_edges = {
        ("input_data", "step_1_loop_1_step_1_python_pandas"): {
            "input_slot_name": "step_1_main_input",
            "output_slot_name": "all",
            "validator": validate_input_file_dummy,
            "env_var": "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
            "filepaths": (
                Path(f"{test_dir}/input_data1/file1.csv"),
                Path(f"{test_dir}/input_data2/file2.csv"),
            ),
        },
        ("step_1_loop_1_step_1_python_pandas", "step_1_2"): {
            "input_slot_name": "step_1_main_input",
            "output_slot_name": "step_1_main_output",
            "validator": validate_input_file_dummy,
            "env_var": "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
            "filepaths": (
                Path("intermediate/step_1_loop_1_step_1_python_pandas/result.parquet"),
            ),
        },
        ("step_1_2", "results"): {
            "input_slot_name": "result",
            "output_slot_name": "step_2_main_output",
            "validator": validate_input_file_dummy,
            "env_var": None,
            "filepaths": (Path("intermediate/step_1_2/result.parquet"),),
        },
    }
    check_nodes_and_edges(pipeline_graph, expected_nodes, expected_edges)


def test_merge_combined_implementations_parallel(
    default_config_params, test_dir, unit_test_specifications_dir
) -> None:
    config_params = default_config_params
    config_params["pipeline"] = load_yaml(
        f"{unit_test_specifications_dir}/pipeline_combine_with_parallel.yaml"
    )
    pipeline_graph = PipelineGraph(Config(config_params))
    expected_nodes = {
        "input_data",
        "step_1_parallel_split_1_step_1_python_pandas",
        "step_1_parallel_split_2_step_1_python_pandas",
        "steps_1_and_2_combined",
        "step_3_python_pandas",
        "step_4_python_pandas",
        "results",
    }
    expected_edges = {
        ("input_data", "step_1_parallel_split_1_step_1_python_pandas"): {
            "input_slot_name": "step_1_main_input",
            "output_slot_name": "all",
            "validator": validate_input_file_dummy,
            "env_var": "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
            "filepaths": (
                Path(f"{test_dir}/input_data1/file1.csv"),
                Path(f"{test_dir}/input_data2/file2.csv"),
            ),
        },
        ("input_data", "step_1_parallel_split_2_step_1_python_pandas"): {
            "input_slot_name": "step_1_main_input",
            "output_slot_name": "all",
            "validator": validate_input_file_dummy,
            "env_var": "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
            "filepaths": (
                Path(f"{test_dir}/input_data1/file1.csv"),
                Path(f"{test_dir}/input_data2/file2.csv"),
            ),
        },
        ("input_data", "steps_1_and_2_combined"): {
            "input_slot_name": "step_1_step_1_main_input",
            "output_slot_name": "all",
            "validator": validate_input_file_dummy,
            "env_var": "STEP_1_DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
            "filepaths": (
                Path(f"{test_dir}/input_data1/file1.csv"),
                Path(f"{test_dir}/input_data2/file2.csv"),
            ),
        },
        ("input_data", "step_4_python_pandas"): {
            "input_slot_name": "step_4_secondary_input",
            "output_slot_name": "all",
            "validator": validate_input_file_dummy,
            "env_var": "DUMMY_CONTAINER_SECONDARY_INPUT_FILE_PATHS",
            "filepaths": (
                Path(f"{test_dir}/input_data1/file1.csv"),
                Path(f"{test_dir}/input_data2/file2.csv"),
            ),
        },
        ("step_1_parallel_split_1_step_1_python_pandas", "steps_1_and_2_combined"): {
            "input_slot_name": "step_2_step_2_main_input",
            "output_slot_name": "step_1_main_output",
            "validator": validate_input_file_dummy,
            "env_var": "STEP_2_DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
            "filepaths": (
                Path(
                    "intermediate/step_1_parallel_split_1_step_1_python_pandas/result.parquet"
                ),
            ),
        },
        ("step_1_parallel_split_2_step_1_python_pandas", "steps_1_and_2_combined"): {
            "input_slot_name": "step_2_step_2_main_input",
            "output_slot_name": "step_1_main_output",
            "validator": validate_input_file_dummy,
            "env_var": "STEP_2_DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
            "filepaths": (
                Path(
                    "intermediate/step_1_parallel_split_2_step_1_python_pandas/result.parquet"
                ),
            ),
        },
        ("steps_1_and_2_combined", "step_3_python_pandas"): {
            "input_slot_name": "step_3_main_input",
            "output_slot_name": "step_2_main_output",
            "validator": validate_input_file_dummy,
            "env_var": "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
            "filepaths": (Path("intermediate/steps_1_and_2_combined/result.parquet"),),
        },
        ("step_3_python_pandas", "step_4_python_pandas"): {
            "input_slot_name": "step_4_main_input",
            "output_slot_name": "step_3_main_output",
            "validator": validate_input_file_dummy,
            "env_var": "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
            "filepaths": (Path("intermediate/step_3_python_pandas/result.parquet"),),
        },
        ("step_4_python_pandas", "results"): {
            "input_slot_name": "result",
            "output_slot_name": "step_4_main_output",
            "validator": validate_input_file_dummy,
            "env_var": None,
            "filepaths": (Path("intermediate/step_4_python_pandas/result.parquet"),),
        },
    }
    check_nodes_and_edges(pipeline_graph, expected_nodes, expected_edges)


@pytest.mark.parametrize(
    "problem_key, error_msg, use_custom_schema",
    [
        (
            "combine_with_iteration_cycle",
            "The pipeline graph contains a cycle after combining implementations: [('step_1_2', 'step_1_loop_2_step_1_python_pandas', 0), "
            "('step_1_loop_2_step_1_python_pandas', 'step_1_2', 0)]",
            True,
        ),
        (
            "combine_with_extra_node",
            "Pipeline configuration nodes ['step_1', 'step_2', 'step_3'] do not match metadata steps ['step_1', 'step_2'].",
            True,
        ),
        (
            "combine_with_missing_node",
            "Pipeline configuration nodes ['step_4'] do not match metadata steps ['step_3', 'step_4'].",
            False,
        ),
        (
            "combine_bad_topology",
            "Pipeline configuration nodes ['step_1b', 'step_1a'] are not topologically consistent with the intended implementations for ['step_1a', 'step_1b']:\nThere is a path from successor step_1b to predecessor step_1a.",
            True,
        ),
        (
            "combine_bad_implementation_names",
            "Pipeline configuration nodes ['step_2', 'step_4'] do not match metadata steps ['step_3', 'step_4'].",
            False,
        ),
    ],
)
def test_bad_combined_configuration_raises(
    problem_key,
    error_msg,
    use_custom_schema,
    default_config_params,
    unit_test_specifications_dir,
) -> None:
    config_params = default_config_params
    config_params["pipeline"] = load_yaml(
        f"{unit_test_specifications_dir}/pipeline_{problem_key}.yaml"
    )
    with pytest.raises(ValueError, match=re.escape(error_msg)):
        if use_custom_schema:
            schema = PipelineSchema(problem_key, *TESTING_SCHEMA_PARAMS[problem_key])
            PipelineGraph(Config(config_params, schema))
        else:
            PipelineGraph(Config(config_params))


def test_nested_templated_steps(
    default_config_params, test_dir, unit_test_specifications_dir
) -> None:
    config_params = default_config_params
    config_params["pipeline"] = load_yaml(
        f"{unit_test_specifications_dir}/pipeline_nested_templated_steps.yaml"
    )
    # Need a custom schema to allow nested TemplatedSteps
    schema = PipelineSchema(
        "nested_templated_steps", *TESTING_SCHEMA_PARAMS["nested_templated_steps"]
    )
    # Ensure that Config instantiates without raising an exception
    pipeline_graph = PipelineGraph(Config(config_params, schema))
    expected_nodes = {
        "input_data",
        "step_1_loop_1_step_1_loop_1_parallel_split_1_step_1_python_pandas",
        "step_1_loop_1_step_1_loop_1_parallel_split_2_step_1_python_pandas",
        "step_1_loop_2_step_1_loop_2_parallel_split_1_step_1_python_pandas",
        "step_1_loop_3_step_1_loop_3_step_1a_step_1a_python_pandas",
        "step_1_loop_3_step_1_loop_3_step_1b_step_1b_python_pandas",
        "results",
    }
    expected_edges = {
        ("input_data", "step_1_loop_1_step_1_loop_1_parallel_split_1_step_1_python_pandas"): {
            "input_slot_name": "step_1_main_input",
            "output_slot_name": "all",
            "validator": validate_input_file_dummy,
            "env_var": "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
            "filepaths": (
                Path(f"{test_dir}/input_data1/file1.csv"),
                Path(f"{test_dir}/input_data2/file2.csv"),
            ),
        },
        ("input_data", "step_1_loop_1_step_1_loop_1_parallel_split_2_step_1_python_pandas"): {
            "input_slot_name": "step_1_main_input",
            "output_slot_name": "all",
            "validator": validate_input_file_dummy,
            "env_var": "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
            "filepaths": (
                Path(f"{test_dir}/input_data1/file1.csv"),
                Path(f"{test_dir}/input_data2/file2.csv"),
            ),
        },
        (
            "step_1_loop_1_step_1_loop_1_parallel_split_1_step_1_python_pandas",
            "step_1_loop_2_step_1_loop_2_parallel_split_1_step_1_python_pandas",
        ): {
            "input_slot_name": "step_1_main_input",
            "output_slot_name": "step_1_main_output",
            "validator": validate_input_file_dummy,
            "env_var": "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
            "filepaths": (
                Path(
                    "intermediate/step_1_loop_1_step_1_loop_1_parallel_split_1_step_1_python_pandas/result.parquet"
                ),
            ),
        },
        (
            "step_1_loop_1_step_1_loop_1_parallel_split_2_step_1_python_pandas",
            "step_1_loop_2_step_1_loop_2_parallel_split_1_step_1_python_pandas",
        ): {
            "input_slot_name": "step_1_main_input",
            "output_slot_name": "step_1_main_output",
            "validator": validate_input_file_dummy,
            "env_var": "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
            "filepaths": (
                Path(
                    "intermediate/step_1_loop_1_step_1_loop_1_parallel_split_2_step_1_python_pandas/result.parquet"
                ),
            ),
        },
        (
            "step_1_loop_2_step_1_loop_2_parallel_split_1_step_1_python_pandas",
            "step_1_loop_3_step_1_loop_3_step_1a_step_1a_python_pandas",
        ): {
            "input_slot_name": "step_1a_main_input",
            "output_slot_name": "step_1_main_output",
            "validator": validate_input_file_dummy,
            "env_var": "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
            "filepaths": (
                Path(
                    "intermediate/step_1_loop_2_step_1_loop_2_parallel_split_1_step_1_python_pandas/result.parquet"
                ),
            ),
        },
        (
            "step_1_loop_3_step_1_loop_3_step_1a_step_1a_python_pandas",
            "step_1_loop_3_step_1_loop_3_step_1b_step_1b_python_pandas",
        ): {
            "input_slot_name": "step_1b_main_input",
            "output_slot_name": "step_1a_main_output",
            "validator": validate_input_file_dummy,
            "env_var": "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
            "filepaths": (
                Path(
                    "intermediate/step_1_loop_3_step_1_loop_3_step_1a_step_1a_python_pandas/result.parquet"
                ),
            ),
        },
        ("step_1_loop_3_step_1_loop_3_step_1b_step_1b_python_pandas", "results"): {
            "input_slot_name": "result",
            "output_slot_name": "step_1b_main_output",
            "validator": validate_input_file_dummy,
            "env_var": None,
            "filepaths": (
                Path(
                    "intermediate/step_1_loop_3_step_1_loop_3_step_1b_step_1b_python_pandas/result.parquet"
                ),
            ),
        },
    }
    check_nodes_and_edges(pipeline_graph, expected_nodes, expected_edges)


####################
# Helper functions #
####################


def check_nodes_and_edges(pipeline_graph, expected_nodes, expected_edges):
    assert set(pipeline_graph.nodes) == set(expected_nodes)
    assert set(pipeline_graph.edges()) == expected_edges.keys()
    for source, sink, edge_attrs in pipeline_graph.edges(data=True):
        assert (
            edge_attrs["input_slot"].name == expected_edges[(source, sink)]["input_slot_name"]
        )
        assert edge_attrs["input_slot"].env_var == expected_edges[(source, sink)]["env_var"]
        assert (
            edge_attrs["input_slot"].validator == expected_edges[(source, sink)]["validator"]
        )
        assert (
            edge_attrs["output_slot"].name
            == expected_edges[(source, sink)]["output_slot_name"]
        )
        assert edge_attrs["filepaths"] == tuple(
            [str(file) for file in expected_edges[(source, sink)]["filepaths"]]
        )



"""
=========
Filepaths
=========
"""

from pathlib import Path

# TODO: We'll need to update this to be more generic for external users and have a way of configuring this
CONTAINER_DIR = "/mnt/team/simulation_science/priv/engineering/er_ecosystem/images"
IMPLEMENTATION_METADATA = Path(__file__).parent.parent / "implementation_metadata.yaml"
# Bind EasyLink temp dir to /tmp in the container.
# For now, put slurm in /tmp to avoid creating a subdir with a prolog script
EASYLINK_TEMP = {"local": Path("/tmp/easylink"), "slurm": Path("/tmp")}

SPARK_SNAKEFILE = Path(__file__).parent / "spark.smk"



import os
import shutil
import subprocess
from pathlib import Path

# This script just tests that the dummy containers can actually be iterated (each taking as input the
# output of the previous). It is not part of the dummy containers themselves!


# https://stackoverflow.com/a/185941/
def rm_in_directory(directory):
    for filename in os.listdir(directory):
        file_path = os.path.join(directory, filename)
        try:
            if os.path.isfile(file_path) or os.path.islink(file_path):
                os.unlink(file_path)
            elif os.path.isdir(file_path):
                shutil.rmtree(file_path)
        except Exception as e:
            print("Failed to delete %s. Reason: %s" % (file_path, e))


user = os.environ["USER"]
singularity_tmp = f"/tmp/singularity_{user}/tmp"
singularity_workdir = f"/tmp/singularity_{user}/workdir"
results_dir = f"/tmp/singularity_{user}/results"
subprocess.run(["mkdir", "-p", singularity_tmp])
subprocess.run(["mkdir", "-p", singularity_workdir])
subprocess.run(["mkdir", "-p", results_dir])
rm_in_directory(singularity_tmp)
rm_in_directory(singularity_workdir)
rm_in_directory(results_dir)

input_file = os.getenv("DUMMY_CONTAINERS_TEST_INPUT_FILE", "./input_data/input_file_2.csv")
input_files = [str(Path(input_file).resolve())]
input_file_format = input_file.split(".")[-1]

# Useful for debugging: you don't need to rebuild the containers for script changes if you use this flag.
use_latest_scripts = os.getenv(
    "DUMMY_CONTAINERS_TEST_USE_LATEST_SCRIPTS", "false"
).lower() in ("true", "1", "t")

num_steps = 30

import random

random.seed(1234)

implementations = [
    "python_pandas/python_pandas.sif",
    "r/r-image.sif",
    "python_pyspark/python_pyspark.sif",
]

steps = [
    {
        "implementation": random.choice(implementations),
        "output_file_format": random.choice(["csv", "parquet"]),
    }
    for _ in range(num_steps)
]

files_so_far = input_files.copy()

for i, step in enumerate(steps):
    print(step)

    step_results_dir = f"{results_dir}/step_{i}"
    subprocess.run(["mkdir", "-p", step_results_dir])

    implementation = step["implementation"]
    output_file_format = step["output_file_format"]

    if (
        os.getenv("DUMMY_CONTAINERS_TEST_INCLUDE_EXTRA_OUTPUTS", "false").lower()
        in ("true", "1", "t")
        and random.random() < 0.25
    ):
        output_files = [
            f"{step_results_dir}/result_1.{output_file_format}",
            f"{step_results_dir}/result_2.{output_file_format}",
        ]
    else:
        output_files = [f"{step_results_dir}/result.{output_file_format}"]

    input_env_vars = ["DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS"]
    env_var_args = {
        "SPARK_MASTER_URL": "local[2]",
        "OUTPUT_FILE_FORMAT": output_file_format,
        "OUTPUT_PATHS": ",".join(
            [
                output_file.replace(step_results_dir, "/results")
                for output_file in output_files
            ]
        ),
    }

    if (
        os.getenv("DUMMY_CONTAINERS_TEST_INCLUDE_BROKEN", "false").lower()
        in ("true", "1", "t")
        and random.random() < 0.25
    ):
        env_var_args["BROKEN"] = "yes"

    if os.getenv("DUMMY_CONTAINERS_TEST_INCLUDE_INCREMENT", "false").lower() in (
        "true",
        "1",
        "t",
    ):
        env_var_args["INCREMENT"] = int(random.random() * 3) + 1

    bindings = [
        (step_results_dir, "/results"),
        (singularity_tmp, "/tmp"),
    ]
    if use_latest_scripts:
        if implementation == "python_pyspark":
            bindings.append(("./python_pyspark/dummy_step.py", "/code/dummy_step.py"))
        elif implementation == "python_pandas":
            bindings.append(("./python_pandas/dummy_step.py", "/dummy_step.py"))
        elif implementation == "r":
            bindings.append(("./r/dummy_step.R", "/dummy_step.R"))
        else:
            raise ValueError()

    bindings.extend([(input_file, input_file) for input_file in input_files])
    env_var_args["MAIN_INPUT_FILE_PATHS"] = ",".join(input_files)

    if (
        os.getenv("DUMMY_CONTAINERS_TEST_INCLUDE_EXTRA_INPUTS", "false").lower()
        in ("true", "1", "t")
        and random.random() < 0.1
    ):
        second_main_input_file = random.choice(files_so_far)
        second_main_input_file_format = second_main_input_file.split(".")[-1]
        env_var_args["MAIN_INPUT_FILE_PATHS"] += "," + second_main_input_file
        bindings.append((second_main_input_file, second_main_input_file))

    if (
        os.getenv("DUMMY_CONTAINERS_TEST_INCLUDE_EXTRA_INPUTS", "false").lower()
        in ("true", "1", "t")
        and random.random() < 0.1
    ):
        secondary_input_file = random.choice(files_so_far)
        secondary_input_file_format = secondary_input_file.split(".")[-1]
        env_var_args["SECONDARY_INPUT_FILE_PATHS"] = secondary_input_file
        input_env_vars.append("DUMMY_CONTAINER_SECONDARY_INPUT_FILE_PATHS")
        bindings.append((secondary_input_file, secondary_input_file))

    if (
        os.getenv("DUMMY_CONTAINERS_TEST_INCLUDE_EXTRA_INPUTS", "false").lower()
        in ("true", "1", "t")
        and random.random() < 0.1
    ):
        implementation_specific_input_file = random.choice(files_so_far)
        implementation_specific_input_file_format = implementation_specific_input_file.split(
            "."
        )[-1]
        env_var_args[
            "EXTRA_IMPLEMENTATION_SPECIFIC_INPUT_FILE_PATH"
        ] = implementation_specific_input_file
        bindings.append(
            (implementation_specific_input_file, implementation_specific_input_file)
        )

    if os.getenv("DUMMY_CONTAINERS_TEST_INCLUDE_EXTRA_OUTPUTS", "false").lower() in (
        "true",
        "1",
        "t",
    ):
        bindings.append((step_results_dir, "/diagnostics/"))

    env_var_args = {f"DUMMY_CONTAINER_{k}": v for k, v in env_var_args.items()}

    if implementation == "python_pyspark":
        bindings.append((singularity_workdir, "/workdir"))
        workdir = "/workdir"
    else:
        workdir = "/"

    env = {**os.environ}

    image = str(Path(f"./{implementation}").resolve())
    command = [
        "singularity",
        "run",
        "--pwd",
        workdir,
        "-B",
        ",".join([f"{k}:{v}" for k, v in bindings]),
        image,
    ]
    env = {**env, **{f"SINGULARITYENV_{k}": v for k, v in env_var_args.items()}}
    env["INPUT_ENV_VARS"] = ",".join(input_env_vars)

    print(" ".join(command))
    print(env_var_args)
    subprocess.check_output(command, env=env)

    files_so_far += output_files
    input_files = output_files
    input_file_format = output_file_format

    # Just to ensure that steps aren't communicating in a way they shouldn't
    rm_in_directory(singularity_tmp)
    rm_in_directory(singularity_workdir)



"""
================
Graph Components
================

This module is responsible for defining the modular building-block objects that 
can be composed to create graph representations of pipelines.

"""

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Callable
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

import networkx as nx

if TYPE_CHECKING:
    from easylink.implementation import Implementation
    from easylink.step import Step


@dataclass(frozen=True)
class InputSlot:
    """A single input slot to a specific node.

    ``InputSlots`` represent distinct semantic categories of input files, between
    which a node must be able to differentiate. In order to pass data between nodes,
    an ``InputSlot`` of one node can be connected to an :class:`OutputSlot` of another
    node via an :class:`EdgeParams` instance.

    Notes
    -----
    Nodes can be either :class:`Steps<easylink.step.Step>` or :class:`Implementations<easylink.implementation.Implementation>`.
    """

    name: str
    """The name of the ``InputSlot``."""
    env_var: str | None
    """The environment variable that is used to pass a list  of data filepaths to 
    an ``Implementation``."""
    validator: Callable[[str], None]
    """A function that validates the input data being passed into the pipeline via 
    this ``InputSlot``. If the data is invalid, the function should raise an exception 
    with a descriptive error message which will then be reported to the user.
    **Note that the function *must* be defined in the** :mod:`easylink.utilities.validation_utils` 
    **module!**"""
    splitter: Callable[[list[str], str, Any], None] | None = None
    """A function that splits the incoming data to this ``InputSlot`` into smaller
    pieces. The primary purpose of this functionality is to run sections of the 
    pipeline in an embarrassingly parallel manner. **Note that the function *must* 
    be defined in the **:mod:`easylink.utilities.splitter_utils`** module!**"""


@dataclass(frozen=True)
class OutputSlot:
    """A single output slot from a specific node.

    ``Outputslots`` represent distinct semantic categories of output files, between
    which a node must be able to differentiate. In order to pass data between nodes,
    an ``OutputSlot`` of one node can be connected to an :class:`InputSlot` of another
    node via an :class:`EdgeParams` instance.

    Notes
    -----
    Nodes can be either :class:`Steps<easylink.step.Step>` or :class:`Implementations<easylink.implementation.Implementation>`.

    Input data is validated via the ``InputSlot`` required :attr:`~InputSlot.validator`
    attribute. In order to prevent multiple validations of the same files (since
    outputs of one node can be inputs to another), no such validator is stored
    here on the ``OutputSlot``.
    """

    name: str
    """The name of the ``OutputSlot``."""
    aggregator: Callable[[list[str], str], None] = None
    """A function that aggregates all of the generated data to be passed out via this
    ``OutputSlot``. The primary purpose of this functionality is to run sections
    of the pipeline in an embarrassingly parallel manner. **Note that the function 
    *must* be defined in the **:py:mod:`easylink.utilities.aggregator_utils`** module!**"""


@dataclass(frozen=True)
class EdgeParams:
    """The details of an edge between two nodes in a graph.

    ``EdgeParams`` connect the :class:`OutputSlot` of a source node to the :class:`InputSlot`
    of a target node.

    Notes
    -----
    Nodes can be either :class:`Steps<easylink.step.Step>` or :class:`Implementations<easylink.implementation.Implementation>`.
    """

    source_node: str
    """The name of the source node."""
    target_node: str
    """The name of the target node."""
    output_slot: str
    """The name of the source node's ``OutputSlot``."""
    input_slot: str
    """The name of the target node's ``InputSlot``."""
    filepaths: tuple[str] | None = None
    """The filepaths that are passed from the source node to the target node."""

    @classmethod
    def from_graph_edge(
        cls: type["EdgeParams"],
        source: str,
        sink: str,
        edge_attrs: dict[str, OutputSlot | InputSlot | str | None],
    ) -> EdgeParams:
        """A convenience method to create an ``EdgeParams`` instance.

        Parameters
        ----------
        source
            The name of the source node.
        sink
            The name of the target node.
        edge_attrs
            The attributes of the edge connecting the source and target nodes.
            'output_slot' and 'input_slot' are required keys while 'filepaths' is
            optional.
        """
        return cls(
            source,
            sink,
            edge_attrs["output_slot"].name,
            edge_attrs["input_slot"].name,
            edge_attrs.get("filepaths"),
        )


class StepGraph(nx.MultiDiGraph):
    """A directed acyclic graph (DAG) of :class:`Steps<easylink.step.Step>`.

    ``StepGraphs`` are DAGs with ``Step`` names for nodes and their corresponding
    ``Step`` instances as attributes on those nodes. The file dependencies between
    nodes are the graph edges; multiple edges between nodes are permitted.

    Notes
    -----
    These are high-level abstractions; they represent a conceptual pipeline
    graph with no detail as to how each ``Step`` is implemented.

    The largest ``StepGraph`` is that of the entire :class:`~easylink.pipeline_schema.PipelineSchema`.
    """

    @property
    def step_nodes(self) -> list[str]:
        """The topologically sorted list of ``Step`` names."""
        ordered_nodes = list(nx.topological_sort(self))
        return [node for node in ordered_nodes if node != "input_data" and node != "results"]

    @property
    def steps(self) -> list[Step]:
        """The topologically sorted list of all ``Steps`` in the graph."""
        return [self.nodes[node]["step"] for node in self.step_nodes]

    def add_node_from_step(self, step: Step) -> None:
        """Adds a new node to the ``StepGraph``.

        Parameters
        ----------
        step
            The ``Step`` to add to the graph as a new node.
        """
        self.add_node(step.name, step=step)

    def add_edge_from_params(self, edge_params: EdgeParams) -> None:
        """Adds a new edge to the ``StepGraph``.

        Parameters
        ----------
        edge_params
            The details of the new edge to be added to the graph.
        """
        return self.add_edge(
            edge_params.source_node,
            edge_params.target_node,
            output_slot=self.nodes[edge_params.source_node]["step"].output_slots[
                edge_params.output_slot
            ],
            input_slot=self.nodes[edge_params.target_node]["step"].input_slots[
                edge_params.input_slot
            ],
        )


class ImplementationGraph(nx.MultiDiGraph):
    """A directed acyclic graph (DAG) of :class:`Implementations<easylink.implementation.Implementation>`.

    ``ImplementationGraphs`` are DAGs with ``Implementations`` for nodes and the
    file dependencies between them for edges. Self-edges as well as multiple edges
    between nodes are permitted.

    Notes
    -----
    An ``ImplementationGraph`` is a low-level abstraction; it represents the *actual
    implementations* of each :class:`~easylink.step.Step` in a pipeline. This is
    in contrast to a :class:`StepGraph`, which can be an intricate nested structure
    due to the various complex and self-similar ``Step`` instances (which represent
    abstract operations such as "loop this step N times"). An ``ImplementationGraph``
    is the flattened and concrete graph of ``Implementations`` in a given pipeline.

    The largest ``ImplementationGraph`` (that is, the specific ``ImplementationGraph``
    representing the entire pipeline to be run) is that of the :class:`~easylink.pipeline_graph.PipelineGraph`.
    """

    @property
    def implementation_nodes(self) -> list[str]:
        """The topologically sorted list of ``Implementation`` names."""
        ordered_nodes = list(nx.topological_sort(self))
        return [node for node in ordered_nodes if node != "input_data" and node != "results"]

    @property
    def implementations(self) -> list[Implementation]:
        """The topologically sorted  list of all ``Implementations`` in the graph."""
        return [self.nodes[node]["implementation"] for node in self.implementation_nodes]

    def add_node_from_implementation(self, node_name, implementation: Implementation) -> None:
        """Adds a new node to the ``ImplementationGraph``.

        Parameters
        ----------
        node_name
            The name of the new node.
        implementation
            The ``Implementation`` to add to the graph as a new node.
        """
        self.add_node(node_name, implementation=implementation)

    def add_edge_from_params(self, edge_params: EdgeParams) -> None:
        """Adds a new edge to the ``ImplementationGraph``.

        Parameters
        ----------
        edge_params
            The details of the new edge to be added to the graph.
        """
        return self.add_edge(
            edge_params.source_node,
            edge_params.target_node,
            output_slot=self.nodes[edge_params.source_node]["implementation"].output_slots[
                edge_params.output_slot
            ],
            input_slot=self.nodes[edge_params.target_node]["implementation"].input_slots[
                edge_params.input_slot
            ],
            filepaths=edge_params.filepaths,
        )


@dataclass(frozen=True)
class SlotMapping(ABC):
    """A mapping between a slot on a parent node and a slot on one of its child nodes.

    ``SlotMapping`` is an interface intended to be used by concrete :class:`InputSlotMapping`
    and :class:`OutputSlotMapping` classes. It represents a mapping between
    parent and child nodes at different levels of a potentially-nested
    :class:`~easylink.pipeline_schema.PipelineSchema`.

    Notes
    -----
    Nodes can be either :class:`Steps<easylink.step.Step>` or :class:`Implementations<easylink.implementation.Implementation>`.
    """

    parent_slot: str
    """The name of the parent slot."""
    child_node: str
    """The name of the child node."""
    child_slot: str
    """The name of the child slot."""

    @abstractmethod
    def remap_edge(self, edge: EdgeParams) -> EdgeParams:
        """Remaps an edge to connect the parent and child nodes."""
        pass


class InputSlotMapping(SlotMapping):
    """A mapping between :class:`InputSlots<InputSlot>` of a parent node and a child node."""

    def remap_edge(self, edge: EdgeParams) -> EdgeParams:
        """Remaps an edge's ``InputSlot``.

        Parameters
        ----------
        edge
            The edge to remap.

        Returns
        -------
            The details of the remapped edge.

        Raises
        ------
        ValueError
            If the parent slot does not match the input slot of the edge.
        """
        if edge.input_slot != self.parent_slot:
            raise ValueError("Parent slot does not match input slot")
        return EdgeParams(
            source_node=edge.source_node,
            target_node=self.child_node,
            output_slot=edge.output_slot,
            input_slot=self.child_slot,
        )


class OutputSlotMapping(SlotMapping):
    """A mapping between :class:`InputSlots<InputSlot>` of a parent node and a child node."""

    def remap_edge(self, edge: EdgeParams) -> EdgeParams:
        """Remaps an edge's :class:`OutputSlot`.

        Parameters
        ----------
        edge
            The edge to remap.

        Returns
        -------
            The details of the remapped edge.

        Raises
        ------
        ValueError
            If the parent slot does not match the output slot of the edge.
        """
        if edge.output_slot != self.parent_slot:
            raise ValueError("Parent slot does not match output slot")
        return EdgeParams(
            source_node=self.child_node,
            target_node=edge.target_node,
            output_slot=self.child_slot,
            input_slot=edge.input_slot,
        )



# mypy: ignore-errors
"""
======================
Command Line Interface
======================

This module is responsible for defining the command line interface (CLI) for running
the EasyLink utility. We use the :mod:`click` library to define the commands and 
their respective options. All commands are accessible via the ``easylink`` group.

.. _example_cli:

Once installed, you can run a pipeline by typing ``easylink run`` into the command 
line and passing in the paths to both a pipeline specification and an input data 
specification:

.. highlight:: console

::

   $ easylink run -p <PIPELINE-SPECIFICATION> -i <INPUT-DATA-SPECIFICATION>

There are several other optional arguments to ``easylink run`` as well;
for help, use ``easylink run --help``. Notably, there is an optional argument to
change where the results are saved (``-o``) as well as one to configure the computing
environment (``-e``).

Note that a schematic of the pipeline's directed acyclic graph (DAG) is automatically 
generated and saved to the output directory as an SVG file which can be rendered 
by most internet browsers. If this schematic is desired _without_ actually running 
the pipeline, use ``easylink generate-dag``:

::

   $ easylink generate-dag -p <PIPELINE-SPECIFICATION> -i <INPUT-DATA-SPECIFICATION

As before, refer to ``easylink generate-dag --help`` for information on other options.

.. _end_example_cli:

For usage documentation, see :ref:`cli`.
"""

from collections.abc import Callable

import click
from loguru import logger

from easylink import runner
from easylink.utilities.data_utils import get_results_directory
from easylink.utilities.general_utils import (
    configure_logging_to_terminal,
    handle_exceptions,
)

SHARED_OPTIONS = [
    click.option(
        "-p",
        "--pipeline-specification",
        required=True,
        type=click.Path(exists=True, dir_okay=False, resolve_path=True),
        help="The path to the pipeline specification YAML file.",
    ),
    click.option(
        "-i",
        "--input-data",
        required=True,
        type=click.Path(exists=True, dir_okay=False, resolve_path=True),
        help=(
            "The path to the input data specification YAML file (not the paths to "
            "the input data themselves)."
        ),
    ),
    click.option(
        "-o",
        "--output-dir",
        type=click.Path(exists=False, dir_okay=True, resolve_path=True),
        help=(
            "The directory to write results and incidental files (logs, etc.) to. "
            "If no value is passed, results will be written to a 'results/' directory "
            "in the current working directory."
        ),
    ),
    click.option(
        "--no-timestamp",
        is_flag=True,
        default=False,
        help="Do not save the results in a timestamped sub-directory of ``--output-dir``.",
    ),
    click.option(
        "-v", "--verbose", count=True, help="Increase logging verbosity.", hidden=True
    ),
    click.option(
        "--pdb",
        "with_debugger",
        is_flag=True,
        help="Drop into python debugger if an error occurs.",
        hidden=True,
    ),
]


def _pass_shared_options(func: Callable) -> Callable:
    """Passes shared options to a click command.

    This function is a decorator that takes a click command callable and adds the
    shared options defined in ``SHARED_OPTIONS`` to it.

    Parameters
    ----------
    func
        The click command function to add shared options to.

    Returns
    -------
        The click command function with the shared options added.
    """
    for option in SHARED_OPTIONS:
        func = option(func)
    return func


@click.group()
def easylink():
    """The command line entrypoint to the EasyLink utility."""
    pass


@easylink.command()
@_pass_shared_options
@click.option(
    "-e",
    "--computing-environment",
    default=None,
    show_default=True,
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
    help=(
        "Path to the computing environment specification YAML file. If no value is passed, "
        "the pipeline will be run locally."
    ),
)
def run(
    pipeline_specification: str,
    input_data: str,
    output_dir: str | None,
    no_timestamp: bool,
    computing_environment: str | None,
    verbose: int,
    with_debugger: bool,
) -> None:
    """Runs a pipeline from the command line.

    In addition to running the pipeline, this command will also generate the directed
    acyclic graph (DAG) image. If you only want to generate the image without actually
    running the pipeline, use the ``easylink generate-dag`` command.
    """
    configure_logging_to_terminal(verbose)
    logger.info("Running pipeline")
    results_dir = get_results_directory(output_dir, no_timestamp).as_posix()
    logger.info(f"Results directory: {results_dir}")
    # TODO [MIC-4493]: Add configuration validation

    main = handle_exceptions(
        func=runner.main, exceptions_logger=logger, with_debugger=with_debugger
    )
    main(
        command="run",
        pipeline_specification=pipeline_specification,
        input_data=input_data,
        computing_environment=computing_environment,
        results_dir=results_dir,
    )
    logger.info("*** FINISHED ***")


@easylink.command()
@_pass_shared_options
def generate_dag(
    pipeline_specification: str,
    input_data: str,
    output_dir: str | None,
    no_timestamp: bool,
    verbose: int,
    with_debugger: bool,
) -> None:
    """Generates an image of the proposed pipeline directed acyclic graph (DAG).

    This command only generates the DAG image of the pipeline; it does not actually
    run it. To run the pipeline, use the ``easylink run`` command.
    """
    configure_logging_to_terminal(verbose)
    logger.info("Generating DAG")
    results_dir = get_results_directory(output_dir, no_timestamp).as_posix()
    logger.info(f"Results directory: {results_dir}")
    # TODO [MIC-4493]: Add configuration validation
    main = handle_exceptions(
        func=runner.main, exceptions_logger=logger, with_debugger=with_debugger
    )
    main(
        command="generate_dag",
        pipeline_specification=pipeline_specification,
        input_data=input_data,
        computing_environment=None,
        results_dir=results_dir,
    )
    logger.info("*** DAG saved to result directory ***")



# mypy: ignore-errors
import pytest


@pytest.mark.skip(
    reason="TODO: MIC-4914: Test the cli tool for `easylink run` to ensure UI doesn't change unexpectedly."
)
def test_easylink_cli():
    pass



# mypy: ignore-errors
import csv
import shutil
from pathlib import Path

import pytest
import yaml

from easylink.configuration import Config, load_params_from_specification


@pytest.fixture(scope="session")
def unit_test_specifications_dir() -> Path:
    return Path(__file__).parent.parent / "specifications" / "unit"


INPUT_DATA_FORMAT_DICT = {
    "correct_cols": [["foo", "bar", "counter"], [1, 2, 3]],
    "wrong_cols": [["wrong", "column", "names"], [1, 2, 3]],
}


def _write_csv(filepath: str, rows: list) -> None:
    with open(filepath, "w") as file:
        writer = csv.writer(file)
        writer.writerows(rows)


@pytest.fixture(scope="session")
def test_dir(tmpdir_factory, unit_test_specifications_dir) -> str:
    """Set up a persistent test directory with some of the specification files"""
    tmp_path = tmpdir_factory.getbasetemp()

    # good pipeline.yaml
    shutil.copy(unit_test_specifications_dir / "pipeline.yaml", tmp_path / "pipeline.yaml")

    # dummy environment.yaml
    shutil.copy(
        unit_test_specifications_dir / "environment_minimum.yaml",
        tmp_path / "environment.yaml",
    )
    shutil.copy(
        unit_test_specifications_dir / "environment_spark_slurm.yaml",
        tmp_path / "spark_environment.yaml",
    )

    # input files
    input_dir1 = tmp_path.mkdir("input_data1")
    input_dir2 = tmp_path.mkdir("input_data2")
    for input_dir in [input_dir1, input_dir2]:
        for base_file in ["file1", "file2"]:
            # good input files
            _write_csv(input_dir / f"{base_file}.csv", INPUT_DATA_FORMAT_DICT["correct_cols"])
            # bad input files
            _write_csv(
                input_dir / f"broken_{base_file}.csv",
                INPUT_DATA_FORMAT_DICT["wrong_cols"],
            )
            # files with wrong extensions
            _write_csv(
                input_dir / f"{base_file}.oops",
                INPUT_DATA_FORMAT_DICT["correct_cols"],
            )

    # good input_data.yaml
    with open(f"{tmp_path}/input_data.yaml", "w") as file:
        yaml.dump(
            {
                "file1": str(input_dir1 / "file1.csv"),
                "file2": str(input_dir2 / "file2.csv"),
            },
            file,
            sort_keys=False,
        )
    # input data is just a list (does not have keys)
    with open(f"{tmp_path}/input_data_list.yaml", "w") as file:
        yaml.dump(
            [
                str(input_dir1 / "file1.csv"),
                str(input_dir2 / "file2.csv"),
            ],
            file,
            sort_keys=False,
        )
    # missing input_data.yaml
    with open(f"{tmp_path}/missing_input_data.yaml", "w") as file:
        yaml.dump(
            {
                "file1": str(input_dir1 / "missing_file1.csv"),
            },
            file,
            sort_keys=False,
        )
    # input directs to files without sensible data
    with open(f"{tmp_path}/bad_columns_input_data.yaml", "w") as file:
        yaml.dump(
            {
                "file1": str(input_dir1 / "broken_file1.csv"),
                "file2": str(input_dir2 / "broken_file2.csv"),
            },
            file,
            sort_keys=False,
        )
    # incorrect file type
    with open(f"{tmp_path}/bad_type_input_data.yaml", "w") as file:
        yaml.dump(
            {
                "file1": str(input_dir1 / "file1.oops"),
                "file2": str(input_dir2 / "file2.oops"),
            },
            file,
            sort_keys=False,
        )

    # bad implementation
    bad_step_implementation_dir = tmp_path.join("steps/foo/implementations/bar/").ensure(
        dir=True
    )
    with open(f"{bad_step_implementation_dir}/metadata.yaml", "w") as file:
        yaml.dump(
            {
                "step": "foo",  # not a supported step
                "image": {
                    "path": "some/path/to/container/directory",
                    "filename": "bar",
                },
            },
            file,
            sort_keys=False,
        )

    return str(tmp_path)


@pytest.fixture()
def results_dir(test_dir) -> Path:
    return Path(f"{test_dir}/results_dir")


@pytest.fixture()
def default_config_paths(test_dir, results_dir) -> dict[str, Path]:
    return {
        "pipeline_specification": Path(f"{test_dir}/pipeline.yaml"),
        "input_data": Path(f"{test_dir}/input_data.yaml"),
        "computing_environment": Path(f"{test_dir}/environment.yaml"),
        "results_dir": results_dir,
    }


@pytest.fixture()
def default_config_params(default_config_paths) -> dict[str, Path]:
    return load_params_from_specification(**default_config_paths)


@pytest.fixture()
def default_config(default_config_params) -> Config:
    """A good/known Config object"""
    return Config(default_config_params)



# mypy: ignore-errors
import os
from pathlib import Path

import pytest

from easylink.utilities.data_utils import (
    copy_configuration_files_to_results_directory,
    create_results_directory,
    create_results_intermediates,
    get_results_directory,
)


def test_create_results_directory(test_dir):
    results_dir = Path(test_dir + "/some/output/dir")
    create_results_directory(results_dir)
    assert results_dir.exists()

    create_results_intermediates(results_dir)
    assert (results_dir / "intermediate").exists()
    assert (results_dir / "diagnostics").exists()


@pytest.mark.parametrize(
    "output_dir_provided, no_timestamp",
    [
        (False, True),
        (False, False),
        (True, True),
        (True, False),
    ],
)
def test_get_results_directory(test_dir, output_dir_provided, no_timestamp, mocker):
    """Tests expected behavior. If directory is provided then a "results/" folder
    is created at the working directory. If no_timestamp is False, then a timestamped
    directory is created within the results directory.
    """
    if output_dir_provided:
        output_dir = test_dir
    else:
        output_dir = None
        # But change the working directory to the test_dir
        os.chdir(test_dir)
    mocker.patch(
        "easylink.utilities.data_utils._get_timestamp", return_value="2024_01_01_00_00_00"
    )
    results_dir = get_results_directory(output_dir, no_timestamp)

    expected_results_dir = Path(test_dir)
    if not output_dir_provided:
        expected_results_dir = expected_results_dir / "results"
    if not no_timestamp:
        expected_results_dir = expected_results_dir / "2024_01_01_00_00_00"

    assert expected_results_dir == results_dir



# mypy: ignore-errors
import errno
import functools
import shutil
import sys
from bdb import BdbQuit
from collections.abc import Callable
from typing import Any, TextIO

import yaml
from loguru import logger


def handle_exceptions(
    func: Callable, exceptions_logger: Any, with_debugger: bool
) -> Callable:
    """Drops a user into an interactive debugger if func raises an error."""

    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except (BdbQuit, KeyboardInterrupt):
            raise
        except Exception as e:
            exceptions_logger.exception("Uncaught exception {}".format(e))
            if with_debugger:
                import pdb
                import traceback

                traceback.print_exc()
                pdb.post_mortem()
            raise

    return wrapped


def configure_logging_to_terminal(verbose: int):
    """Sets up logging to ``sys.stdout``.

    Parameters
    ----------
    verbose
        Verbosity of the logger.
    """
    logger.remove(0)  # Clear default configuration
    _add_logging_sink(sys.stdout, verbose, colorize=True)


def _add_logging_sink(
    sink: TextIO, verbose: int, colorize: bool = False, serialize: bool = False
):
    """Adds a logging sink to the global process logger.

    Parameters
    ----------
    sink
        Either a file or system file descriptor like ``sys.stdout``.
    verbose
        Verbosity of the logger.
    colorize
        Whether to use the colorization options from :mod:`loguru`.
    serialize
        Whether the logs should be converted to JSON before they're dumped
        to the logging sink.

    """
    message_format = (
        "<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | <green>{elapsed}</green> | "
        "<cyan>{function}</cyan>:<cyan>{line}</cyan> - <level>{message}</level>"
    )
    if verbose == 0:
        logger.add(
            sink,
            colorize=colorize,
            level="INFO",
            format=message_format,
            serialize=serialize,
        )
    elif verbose >= 1:
        logger.add(
            sink,
            colorize=colorize,
            level="DEBUG",
            format=message_format,
            serialize=serialize,
        )


def exit_with_validation_error(error_msg: dict) -> None:
    """Logs error messages and exits the program.

    This function logs the provided validation error messages using a structured
    YAML format and terminates the program execution with a non-zero exit code
    (indicating an error).

    Parameters
    ----------
    error_msg
        The error message to print to the user.

    Raises
    ------
    SystemExit
        Exits the program with an EINVAL (invalid argument) code due to
        previously-determined validation errors.

    """

    logger.error(
        "\n\n=========================================="
        "\nValidation errors found. Please see below."
        f"\n\n{yaml.dump(error_msg)}"
        "\nValidation errors found. Please see above."
        "\n==========================================\n"
    )
    sys.exit(errno.EINVAL)


def is_on_slurm() -> bool:
    """Returns True if the current environment is a SLURM cluster."""
    return shutil.which("sbatch") is not None



# mypy: ignore-errors
import os
from pathlib import Path

import pytest

from easylink.configuration import Config, load_params_from_specification
from easylink.pipeline import Pipeline
from easylink.utilities.data_utils import (
    copy_configuration_files_to_results_directory,
    create_results_directory,
    create_results_intermediates,
)

PIPELINE_STRINGS = {
    "local": "rule_strings/pipeline_local.txt",
    "slurm": "rule_strings/pipeline_slurm.txt",
}


@pytest.mark.parametrize("computing_environment", ["local", "slurm"])
def test_build_snakefile(
    default_config_paths, mocker, test_dir, results_dir, computing_environment
):
    config_paths = default_config_paths
    if computing_environment == "slurm":
        config_paths["computing_environment"] = f"{test_dir}/spark_environment.yaml"

    config_params = load_params_from_specification(**config_paths)
    config = Config(config_params)
    mocker.patch("easylink.implementation.Implementation.validate", return_value={})
    pipeline = Pipeline(config)
    create_results_directory(results_dir)
    create_results_intermediates(results_dir)
    copy_configuration_files_to_results_directory(**config_paths)
    snakefile = pipeline.build_snakefile()
    expected_file_path = (
        Path(os.path.dirname(__file__)) / PIPELINE_STRINGS[computing_environment]
    )
    with open(expected_file_path) as expected_file:
        expected = expected_file.read()
    expected = expected.replace("{test_dir}", test_dir)
    snake_str = snakefile.read_text()
    snake_str_lines = snake_str.split("\n")
    expected_lines = expected.split("\n")
    assert len(snake_str_lines) == len(expected_lines)
    for i, expected_line in enumerate(expected_lines):
        assert snake_str_lines[i].strip() == expected_line.strip()



from pathlib import Path
from re import match

import networkx as nx

from easylink.graph_components import InputSlot, OutputSlot
from easylink.pipeline_schema import PIPELINE_SCHEMAS, PipelineSchema
from easylink.pipeline_schema_constants import ALLOWED_SCHEMA_PARAMS
from easylink.step import Step
from easylink.utilities.aggregator_utils import concatenate_datasets
from easylink.utilities.splitter_utils import split_data_by_size
from easylink.utilities.validation_utils import validate_input_file_dummy


def test_schema_instantiation() -> None:
    nodes, edges = ALLOWED_SCHEMA_PARAMS["development"]
    schema = PipelineSchema("development", nodes=nodes, edges=edges)
    sorted_graph = nx.topological_sort(schema.step_graph)
    """Test that the schema is correctly loaded from the pipeline.yaml"""
    assert list(sorted_graph) == [
        "input_data",
        "step_1",
        "step_2",
        "step_3",
        "choice_section",
        "results",
    ]
    step_types = [node["step"] for node in sorted_graph]
    expected_step_types = [type(step) for step in ALLOWED_SCHEMA_PARAMS["development"]]
    for step_type, expected_step_types in zip(step_types, expected_step_types):
        assert isinstance(step_type, expected_step_types)


def test_get_schemas() -> None:
    supported_schemas = PIPELINE_SCHEMAS
    assert isinstance(supported_schemas, list)
    # Ensure list is populated
    assert supported_schemas
    # Check basic structure
    for schema in supported_schemas:
        assert schema.name
        assert schema.step_graph.steps
        assert isinstance(schema.step_graph.steps, list)
        for step in schema.step_graph.steps:
            assert isinstance(step, Step)
            assert step.name


def test_validate_input(test_dir: str) -> None:
    nodes, edges = ALLOWED_SCHEMA_PARAMS["development"]
    schema = PipelineSchema("development", nodes=nodes, edges=edges)
    input_data = {"file1": Path(test_dir) / "input_data1/file1.csv"}
    errors = schema.validate_inputs(input_data)
    assert not errors
    # Test with a bad file
    file_name = Path(test_dir) / "input_data1/broken_file1.csv"
    input_data = {"file1": file_name}
    errors = schema.validate_inputs(input_data)
    assert errors
    assert match(
        "Data file .* is missing required column\\(s\\) .*", errors[str(file_name)][0]
    )


def test_pipeline_schema_get_implementation_graph(default_config) -> None:
    nodes, edges = ALLOWED_SCHEMA_PARAMS["development"]
    schema = PipelineSchema("development", nodes=nodes, edges=edges)
    schema.configure_pipeline(default_config.pipeline, default_config.input_data)
    implementation_graph = schema.get_implementation_graph()
    assert list(implementation_graph.nodes) == [
        "input_data",
        "step_1_python_pandas",
        "step_2_python_pandas",
        "step_3_python_pandas",
        "step_4_python_pandas",
        "results",
    ]
    expected_edges = [
        (
            "input_data",
            "step_1_python_pandas",
            {
                "output_slot": OutputSlot("all"),
                "input_slot": InputSlot(
                    name="step_1_main_input",
                    env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    validator=validate_input_file_dummy,
                ),
                "filepaths": None,
            },
        ),
        (
            "input_data",
            "step_4_python_pandas",
            {
                "output_slot": OutputSlot("all"),
                "input_slot": InputSlot(
                    name="step_4_secondary_input",
                    env_var="DUMMY_CONTAINER_SECONDARY_INPUT_FILE_PATHS",
                    validator=validate_input_file_dummy,
                ),
                "filepaths": None,
            },
        ),
        (
            "step_1_python_pandas",
            "step_2_python_pandas",
            {
                "output_slot": OutputSlot("step_1_main_output"),
                "input_slot": InputSlot(
                    name="step_2_main_input",
                    env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    validator=validate_input_file_dummy,
                ),
                "filepaths": None,
            },
        ),
        (
            "step_2_python_pandas",
            "step_3_python_pandas",
            {
                "output_slot": OutputSlot("step_2_main_output"),
                "input_slot": InputSlot(
                    name="step_3_main_input",
                    env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    validator=validate_input_file_dummy,
                    splitter=split_data_by_size,
                ),
                "filepaths": None,
            },
        ),
        (
            "step_3_python_pandas",
            "step_4_python_pandas",
            {
                "output_slot": OutputSlot(
                    "step_3_main_output", aggregator=concatenate_datasets
                ),
                "input_slot": InputSlot(
                    name="step_4_main_input",
                    env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    validator=validate_input_file_dummy,
                ),
                "filepaths": None,
            },
        ),
        (
            "step_4_python_pandas",
            "results",
            {
                "output_slot": OutputSlot("step_4_main_output"),
                "input_slot": InputSlot(
                    name="result", env_var=None, validator=validate_input_file_dummy
                ),
                "filepaths": None,
            },
        ),
    ]
    assert len(implementation_graph.edges) == len(expected_edges)
    for edge in expected_edges:
        assert edge in implementation_graph.edges(data=True)



from layered_config_tree import LayeredConfigTree

from easylink.graph_components import (
    EdgeParams,
    ImplementationGraph,
    InputSlot,
    InputSlotMapping,
    OutputSlot,
    OutputSlotMapping,
    SlotMapping,
    StepGraph,
)
from easylink.implementation import Implementation
from easylink.step import Step
from easylink.utilities.validation_utils import validate_input_file_dummy


def test_input_slot() -> None:
    input_slot = InputSlot(
        "file1", "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS", validate_input_file_dummy
    )
    assert input_slot.name == "file1"
    assert input_slot.env_var == "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS"
    assert input_slot.validator == validate_input_file_dummy


def test_output_slot() -> None:
    output_slot = OutputSlot("file1")
    assert output_slot.name == "file1"


def test_edge() -> None:
    edge = EdgeParams(
        source_node="input_data",
        target_node="step_1",
        output_slot="file1",
        input_slot="step_1_main_input",
    )
    assert edge.source_node == "input_data"
    assert edge.target_node == "step_1"
    assert edge.output_slot == "file1"
    assert edge.input_slot == "step_1_main_input"


def test_step_graph() -> None:
    step_graph = StepGraph()
    step1 = Step("step_1", output_slots=[OutputSlot("foo")])
    step2 = Step("step_2", input_slots=[InputSlot("bar", None, None)])
    step_graph.add_node_from_step(step1)
    step_graph.add_node_from_step(step2)
    assert step_graph.nodes["step_1"]["step"] == step1
    assert step_graph.nodes["step_2"]["step"] == step2

    edge = EdgeParams(
        source_node="step_1",
        target_node="step_2",
        output_slot="foo",
        input_slot="bar",
    )
    step_graph.add_edge_from_params(edge)
    assert step_graph.edges["step_1", "step_2", 0]["output_slot"].name == "foo"
    assert step_graph.edges["step_1", "step_2", 0]["input_slot"].name == "bar"


def test_implementation_graph(mocker) -> None:
    mocker.patch("easylink.implementation.Implementation._load_metadata")
    mocker.patch("easylink.implementation.Implementation.validate", return_value=[])
    implementation_graph = ImplementationGraph()
    implementation1 = Implementation(
        ["step_1"],
        LayeredConfigTree({"name": "step1"}),
        output_slots=[OutputSlot("foo")],
    )
    implementation2 = Implementation(
        ["step_2"],
        LayeredConfigTree({"name": "step2"}),
        input_slots=[InputSlot("bar", None, None)],
    )
    implementation_graph.add_node_from_implementation("step_1", implementation1)
    implementation_graph.add_node_from_implementation("step_2", implementation2)
    assert implementation_graph.nodes["step_1"]["implementation"] == implementation1
    assert implementation_graph.nodes["step_2"]["implementation"] == implementation2

    edge = EdgeParams(
        source_node="step_1",
        target_node="step_2",
        output_slot="foo",
        input_slot="bar",
    )
    implementation_graph.add_edge_from_params(edge)
    assert implementation_graph.edges["step_1", "step_2", 0]["output_slot"].name == "foo"
    assert implementation_graph.edges["step_1", "step_2", 0]["input_slot"].name == "bar"


def test_input_slot_mapping() -> None:
    input_slot_mapping = InputSlotMapping(
        "step_1_main_input",
        "step_1a",
        "step_1a_main_input",
    )
    assert input_slot_mapping.parent_slot == "step_1_main_input"
    assert input_slot_mapping.child_node == "step_1a"
    assert input_slot_mapping.child_slot == "step_1a_main_input"

    edge = EdgeParams(
        source_node="input_data",
        target_node="step_1",
        output_slot="file1",
        input_slot="step_1_main_input",
    )
    new_edge = input_slot_mapping.remap_edge(edge)
    assert new_edge.source_node == "input_data"
    assert new_edge.target_node == "step_1a"
    assert new_edge.output_slot == "file1"
    assert new_edge.input_slot == "step_1a_main_input"


def test_output_slot_mapping() -> None:
    output_slot_mapping = OutputSlotMapping(
        "step_1_main_output",
        "step_1a",
        "step_1a_main_input",
    )
    assert output_slot_mapping.parent_slot == "step_1_main_output"
    assert output_slot_mapping.child_node == "step_1a"
    assert output_slot_mapping.child_slot == "step_1a_main_input"

    edge = EdgeParams(
        source_node="step_1",
        target_node="output_data",
        output_slot="step_1_main_output",
        input_slot="file1",
    )
    new_edge = output_slot_mapping.remap_edge(edge)
    assert new_edge.source_node == "step_1a"
    assert new_edge.target_node == "output_data"
    assert new_edge.output_slot == "step_1a_main_input"
    assert new_edge.input_slot == "file1"
